import pytest
import torch

from elasticai.creator.arithmetic import FxpArithmetic, FxpParams
from tests.tensor_test_case import assertTensorEqual


@pytest.mark.parametrize("total_bits, frac_bits", [(2, 2), (4, 3), (8, 4)])
def test_cut_float_to_integer_signed(total_bits: int, frac_bits: int) -> None:
    config = FxpArithmetic(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    )
    chck = [
        config.config.minimum_as_integer,
        -1,
        0,
        +1,
        config.config.maximum_as_integer,
    ]
    stimuli = [
        config.minimum_as_rational,
        -config.config.minimum_step_as_rational,
        0.0,
        config.config.minimum_step_as_rational,
        config.maximum_as_rational,
    ]
    rslt = [config.cut_as_integer(val) for val in stimuli]
    assert rslt == chck


@pytest.mark.parametrize("total_bits, frac_bits", [(2, 2), (4, 3), (8, 4)])
def test_cut_tensor_to_integer_signed(total_bits: int, frac_bits: int) -> None:
    config = FxpArithmetic(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    )

    chck = torch.Tensor(
        [config.config.minimum_as_integer, -1, 0, +1, config.config.maximum_as_integer]
    )
    stimuli = torch.Tensor(
        [
            config.config.minimum_as_rational,
            -config.config.minimum_step_as_rational,
            0.0,
            config.config.minimum_step_as_rational,
            config.config.maximum_as_rational,
        ]
    )
    rslt = config.cut_as_integer(stimuli)
    assertTensorEqual(rslt, chck)


@pytest.mark.parametrize("total_bits, frac_bits", [(2, 2), (4, 3), (8, 4)])
def test_cut_x_to_integer_signed(total_bits: int, frac_bits: int) -> None:
    config = FxpArithmetic(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    )

    stimuli_tensor = torch.Tensor(
        [
            config.config.minimum_as_rational,
            -config.config.minimum_step_as_rational,
            0,
            config.config.minimum_step_as_rational,
            config.config.maximum_as_rational,
        ]
    )
    stimuli_float = [
        config.config.minimum_as_rational,
        -config.config.minimum_step_as_rational,
        0.0,
        config.config.minimum_step_as_rational,
        config.config.maximum_as_rational,
    ]
    rslt_tensor = config.cut_as_integer(stimuli_tensor).tolist()
    rslt_float = [config.cut_as_integer(val) for val in stimuli_float]
    assert rslt_tensor == rslt_float


@pytest.mark.parametrize("total_bits, frac_bits", [(2, 2), (4, 3), (8, 4)])
def test_round_float_to_integer_signed(total_bits: int, frac_bits: int) -> None:
    config = FxpArithmetic(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    )
    chck = [
        config.config.minimum_as_integer,
        -1,
        0,
        +1,
        config.config.maximum_as_integer,
    ]
    stimuli = [
        config.config.minimum_as_rational,
        -config.config.minimum_step_as_rational,
        0.0,
        config.config.minimum_step_as_rational,
        config.config.maximum_as_rational,
    ]
    rslt = [config.round_to_integer(val) for val in stimuli]
    assert rslt == chck


@pytest.mark.parametrize("total_bits, frac_bits", [(2, 2), (4, 3), (8, 4)])
def test_round_tensor_to_integer_signed(total_bits: int, frac_bits: int) -> None:
    config = FxpArithmetic(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    )

    chck = torch.Tensor(
        [config.config.minimum_as_integer, -1, 0, +1, config.config.maximum_as_integer]
    )
    stimuli = torch.Tensor(
        [
            config.config.minimum_as_rational,
            -config.config.minimum_step_as_rational,
            0.0,
            config.config.minimum_step_as_rational,
            config.config.maximum_as_rational,
        ]
    )
    rslt = config.round_to_integer(stimuli)
    assertTensorEqual(rslt, chck)


@pytest.mark.parametrize("total_bits, frac_bits", [(2, 2), (4, 3), (8, 4)])
def test_round_x_to_integer_signed(total_bits: int, frac_bits: int) -> None:
    config = FxpArithmetic(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    )

    stimuli_tensor = torch.Tensor(
        [
            config.config.minimum_as_rational,
            -config.config.minimum_step_as_rational,
            0.0,
            config.config.minimum_step_as_rational,
            config.config.maximum_as_rational,
        ]
    )
    stimuli_float = [
        config.config.minimum_as_rational,
        -config.config.minimum_step_as_rational,
        0.0,
        config.config.minimum_step_as_rational,
        config.config.maximum_as_rational,
    ]
    rslt_tensor = config.round_to_integer(stimuli_tensor).tolist()
    rslt_float = [config.round_to_integer(val) for val in stimuli_float]
    assert rslt_tensor == rslt_float


@pytest.mark.parametrize("total_bits, frac_bits", [(2, 2), (4, 3), (8, 4)])
def test_cut_float_to_integer_unsigned(total_bits: int, frac_bits: int) -> None:
    config = FxpArithmetic(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=False)
    )
    chck = [
        config.config.minimum_as_integer,
        1,
        config.config.maximum_as_integer - 1,
        config.config.maximum_as_integer,
    ]
    stimuli = [
        config.config.minimum_as_rational,
        config.config.minimum_step_as_rational,
        config.config.maximum_as_rational - config.config.minimum_step_as_rational,
        config.config.maximum_as_rational,
    ]
    rslt = [config.cut_as_integer(val) for val in stimuli]
    assert rslt == chck


@pytest.mark.parametrize("total_bits, frac_bits", [(2, 2), (4, 3), (8, 4)])
def test_cut_tensor_to_integer_unsigned(total_bits: int, frac_bits: int) -> None:
    config = FxpArithmetic(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=False)
    )
    chck = [
        config.config.minimum_as_integer,
        1,
        config.config.maximum_as_integer - 1,
        config.config.maximum_as_integer,
    ]
    stimuli = torch.Tensor(
        [
            config.config.minimum_as_rational,
            config.config.minimum_step_as_rational,
            config.config.maximum_as_rational - config.config.minimum_step_as_rational,
            config.config.maximum_as_rational,
        ]
    )
    rslt = config.cut_as_integer(stimuli)
    assertTensorEqual(rslt, chck)


@pytest.mark.parametrize("total_bits, frac_bits", [(2, 2), (4, 3), (8, 4)])
def test_round_tensor_to_integer_unsigned(total_bits: int, frac_bits: int) -> None:
    config = FxpArithmetic(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=False)
    )

    chck = [
        config.config.minimum_as_integer,
        1,
        config.config.maximum_as_integer - 1,
        config.config.maximum_as_integer,
    ]
    stimuli = torch.Tensor(
        [
            config.config.minimum_as_rational,
            config.config.minimum_step_as_rational,
            config.config.maximum_as_rational - config.config.minimum_step_as_rational,
            config.config.maximum_as_rational,
        ]
    )
    rslt = config.round_to_integer(stimuli)
    assertTensorEqual(rslt, chck)


@pytest.mark.parametrize(
    "total_bits, frac_bits, signed, val_in, val_out",
    [
        (2, 1, True, 1, 0.5),
        (2, 1, True, 0, 0.0),
        (3, 2, True, -4, -1.0),
        (3, 2, True, 3, 0.75),
        (2, 1, False, 1, 0.5),
        (2, 1, False, 0, 0.0),
        (3, 2, False, 4, 1.0),
        (3, 2, False, 7, 1.75),
    ],
)
def test_integer_to_rational(
    total_bits: int, frac_bits: int, signed: bool, val_in: int, val_out: float
) -> None:
    result = FxpArithmetic(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=signed)
    ).as_rational(val_in)
    assert result == val_out



from collections.abc import Callable, Iterable
from typing import Any

import torch.nn as nn
from torch import fx

from elasticai.creator.ir import Attribute, Edge, GraphProtocol, LoweringPass, Node
from elasticai.creator.ir import Implementation as _Implementation

from .default_handlers import handlers


class Implementation(_Implementation[Node, Edge]):
    name: str
    type: str

    def __init__(
        self,
        *,
        graph: GraphProtocol,
        name: str | None = None,
        type: str | None = None,
        data: dict[str, Attribute] | None = None,
    ):
        super().__init__(data=data, graph=graph, node_fn=Node, edge_fn=Edge)
        if name is not None:
            self.name = name
        if type is not None:
            self.type = type


class Ir2Torch(LoweringPass[Implementation, nn.Module]):
    def convert(
        self, ir: Iterable[Implementation], state_dict: dict[str, Any] | None = None
    ) -> nn.Module:
        """Rebuild the original pytorch model from a given IR.

        Implemenation names containing dots will result in the corresponding modules sorted
        into a corresponding object hierarchy. E.g., for the implementation
        name `'top.linear'` we will create a pytorch container module under the name
        `'top'` and add the linear layer to it under the name `'linear'`. Note that this
        is an implementation detail of Ir2Torch and not a semantic meaning assigned to
        the `'.'` character.

        :param: `ir`: You need to make sure that `Ir2Torch` has a type handler for each implementation in `ir`
        :param: `state_dict`: You can optionally pass a state dict. This should be a state dict created
            from the original model via `nn.Module.state_dict`. As the `Torch2Ir` stage got rid of all
            duplicate submodules, we will strip all unknown keys from the `state_dict` and then load it.
        """
        root_module = nn.Module()
        root = ""
        _ir = {impl.name: impl for impl in ir}
        for impl in _ir.values():
            if impl.type != "module":
                modules = list(self([impl]))
                assert len(modules) == 1
                name = impl.name
                last_parent = root_module
                while "." in name:
                    parent_name, name = name.rsplit(".", 1)
                    last_children = dict(last_parent.named_children())
                    if parent_name not in last_children:
                        current_parent = nn.Module()
                        last_parent.add_module(parent_name, current_parent)
                    else:
                        current_parent = last_children[parent_name]
                    last_parent = current_parent
                last_parent.add_module(name, modules[0])

            else:
                root = impl.name

        graph = fx.Graph()
        ir_root = _ir[root]
        nodes: dict[str, fx.Node] = {}

        def _add_node(ir_node) -> None:
            if ir_node.type not in ("input", "output"):
                predecessors = tuple(
                    nodes[node] for node in ir_root.predecessors(ir_node)
                )
                node = graph.create_node(
                    op="call_module",
                    target=ir_node.data["implementation"],
                    args=predecessors,
                    name=ir_node.name,
                )
                nodes[ir_node.name] = node
            elif ir_node.type == "input":
                n = graph.create_node(
                    op="placeholder", name=ir_node.name, target=ir_node.name
                )
                nodes[ir_node.name] = n
            elif ir_node.type == "output":
                predecessors = tuple(
                    nodes[node] for node in ir_root.predecessors(ir_node)
                )
                n = graph.create_node(
                    op="output",
                    target=ir_node.name,
                    args=predecessors,
                )
                nodes[ir_node.name] = n

        def visit_node(ir_node):
            if ir_node.name not in nodes:
                for pred in ir_root.predecessors(ir_node).values():
                    visit_node(pred)
                if ir_node.name not in nodes:
                    _add_node(ir_node)

        visit_node(ir_root.nodes["output"])

        module = fx.GraphModule(root_module, graph)

        if state_dict is not None:
            filtered_state_dict = {}
            for key, value in state_dict.items():
                submodule = ".".join(key.split(".")[:-1])
                if submodule in _ir:
                    filtered_state_dict[key] = value
            module.load_state_dict(filtered_state_dict)
        return module

    def register_type_handlers(
        self, handlers: Iterable[Callable[[Implementation], nn.Module]]
    ) -> None:
        for handler in handlers:
            self.register(handler.__name__)(handler)


def get_default_converter() -> Ir2Torch:
    converter = Ir2Torch()
    converter.register_type_handlers(handlers)
    return converter



from types import SimpleNamespace
from typing import cast

import torch
from torch import Tensor, tensor

from elasticai.creator.nn.binary.binary_quantization_function import Binarize
from tests.tensor_test_case import TensorTestCase


def binarize(x: Tensor) -> Tensor:
    return cast(Tensor, Binarize.apply(x))


class BinarizeFunctionTest(TensorTestCase):
    def test_ForwardRaisesErrorOnMissingInput(self) -> None:
        def save_for_backward_dummy(_):
            pass

        dummy_context = SimpleNamespace(save_for_backward=save_for_backward_dummy)
        try:
            Binarize.forward(dummy_context)
        except TypeError:
            return

        self.fail()

    def test_Yields1For0(self) -> None:
        self.assertTensorEqual(expected=tensor([1.0]), actual=binarize(tensor([0.0])))

    def test_Yields1For2Point4(self) -> None:
        self.assertTensorEqual(expected=tensor([1.0]), actual=binarize(tensor([2.4])))

    def test_YieldMinus1ForNegativeInput(self) -> None:
        self.assertTensorEqual(expected=tensor([-1.0]), actual=binarize(tensor([-2.8])))

    def check_gradient(self, expected_grad, x) -> None:
        x = torch.tensor([x], requires_grad=True)
        y = binarize(x)
        y.backward()
        self.assertTensorEqual(torch.tensor([expected_grad]), cast(Tensor, x.grad))

    def test_gradient_is_0_for_input_greater_1(self) -> None:
        self.check_gradient(expected_grad=0.0, x=1.1)

    def test_gradient_is_0_for_input_less_minus_1(self) -> None:
        self.check_gradient(expected_grad=0.0, x=-1.1)

    def test_gradient_is_1_for_input_between_minus_1_and_1(self) -> None:
        self.check_gradient(expected_grad=1.0, x=-0.8)



from typing import Any, Callable

from elasticai.creator.ir2vhdl import Instance, LogicVectorSignal, VhdlNode

from .node_factory import (
    InstanceFactoryForCombinatorial,
)


@InstanceFactoryForCombinatorial.register("output")
@InstanceFactoryForCombinatorial.register("input")
@InstanceFactoryForCombinatorial.register("lutron")
@InstanceFactoryForCombinatorial.register
def unclocked_combinatorial(node):
    return UnclockedInstance(node)


class UnclockedInstance(Instance):
    def __init__(
        self,
        node: VhdlNode,
        generic_map: dict[str, Any] | Callable[[], dict[str, Any]] = lambda: {},
    ):
        if callable(generic_map):
            generic_map = generic_map()
        super().__init__(
            node,
            generic_map=node.attributes.get("generic_map", {}) | generic_map,  # pyright: ignore
            port_map=dict(
                d_in=LogicVectorSignal(name="d_in", width=node.input_shape.size()),
                d_out=LogicVectorSignal(name="d_out", width=node.output_shape.size()),
            ),
        )



import pytest

from elasticai.creator.arithmetic.fxp_converter import (
    FxpParams,
)


@pytest.mark.parametrize(
    "total_bits, frac_bits, signed, expected",
    [
        (2, 1, True, -2),
        (2, 1, False, 0),
        (3, 1, True, -4),
        (3, 1, False, 0),
        (5, 1, True, -16),
        (5, 1, False, 0),
    ],
)
def test_fxp_minimum_as_integer(
    total_bits: int, frac_bits: int, signed: bool, expected: int
) -> None:
    result = FxpParams(
        total_bits=total_bits, frac_bits=frac_bits, signed=signed
    ).minimum_as_integer
    assert result == expected


@pytest.mark.parametrize(
    "total_bits, frac_bits, signed, expected",
    [
        (2, 1, True, 1),
        (2, 1, False, 3),
        (3, 1, True, 3),
        (3, 1, False, 7),
        (5, 1, True, 15),
        (5, 1, False, 31),
    ],
)
def test_fxp_maximum_as_integer(
    total_bits: int, frac_bits: int, signed: bool, expected: int
) -> None:
    result = FxpParams(
        total_bits=total_bits, frac_bits=frac_bits, signed=signed
    ).maximum_as_integer
    assert result == expected


@pytest.mark.parametrize(
    "total_bits, frac_bits, signed, expected",
    [
        (2, 0, True, -2.0),
        (2, 1, True, -1.0),
        (2, 1, False, 0.0),
        (3, 1, True, -2.0),
        (3, 1, False, 0.0),
        (5, 1, True, -8.0),
        (5, 1, False, 0.0),
    ],
)
def test_fxp_minimum_as_rational(
    total_bits: int, frac_bits: int, signed: bool, expected: int
) -> None:
    result = FxpParams(
        total_bits=total_bits, frac_bits=frac_bits, signed=signed
    ).minimum_as_rational
    assert result == expected


@pytest.mark.parametrize(
    "total_bits, frac_bits, signed, expected",
    [
        (2, 1, True, 0.5),
        (2, 2, True, 0.25),
        (2, 2, False, 0.25),
        (3, 1, True, 0.5),
        (3, 2, True, 0.25),
        (3, 2, False, 0.25),
        (5, 3, True, 0.125),
        (5, 4, False, 0.0625),
    ],
)
def test_fxp_minimum_step_as_rational(
    total_bits: int, frac_bits: int, signed: bool, expected: int
) -> None:
    result = FxpParams(
        total_bits=total_bits, frac_bits=frac_bits, signed=signed
    ).minimum_step_as_rational
    assert result == expected


@pytest.mark.parametrize(
    "total_bits, frac_bits, signed, expected",
    [
        (2, 0, True, 1.0),
        (2, 1, True, 0.5),
        (2, 1, False, 1.5),
        (3, 2, True, 0.75),
        (3, 2, False, 1.75),
        (5, 2, True, 3.75),
        (5, 3, False, 3.875),
    ],
)
def test_fxp_maximum_as_rational(
    total_bits: int, frac_bits: int, signed: bool, expected: int
) -> None:
    result = FxpParams(
        total_bits=total_bits, frac_bits=frac_bits, signed=signed
    ).maximum_as_rational
    assert result == expected


@pytest.mark.parametrize(
    "total_bits, frac_bits, signed, val_in, expected",
    [
        (2, 1, True, 2, True),
        (2, 1, True, 1, False),
        (3, 1, True, 4, True),
        (3, 1, True, 3, False),
        (2, 1, False, 4, True),
        (2, 1, False, 3, False),
        (3, 1, False, 8, True),
        (3, 1, False, 7, False),
    ],
)
def test_integer_upper_limit(
    total_bits: int, frac_bits: int, signed: bool, val_in: int, expected: bool
) -> None:
    result = FxpParams(
        total_bits=total_bits, frac_bits=frac_bits, signed=signed
    ).integer_out_overflow(val_in)
    assert result == expected


@pytest.mark.parametrize(
    "total_bits, frac_bits, signed, val_in, expected",
    [
        (2, 1, True, -3, True),
        (2, 1, True, -2, False),
        (3, 1, True, -5, True),
        (3, 1, True, -4, False),
        (2, 1, False, -1, True),
        (2, 1, False, 0, False),
        (3, 1, False, -1, True),
        (3, 1, False, 0, False),
    ],
)
def test_integer_downer_limit(
    total_bits: int, frac_bits: int, signed: bool, val_in: int, expected: bool
) -> None:
    result = FxpParams(
        total_bits=total_bits, frac_bits=frac_bits, signed=signed
    ).integer_out_underflow(val_in)
    assert result == expected


@pytest.mark.parametrize(
    "total_bits, frac_bits, signed, val_in, expected",
    [
        (2, 1, True, -3, True),
        (2, 1, True, 2, True),
        (2, 1, True, -2, False),
        (2, 1, True, 1, False),
        (3, 1, True, -5, True),
        (3, 1, True, -4, False),
        (3, 1, True, 3, False),
        (3, 1, True, 4, True),
        (2, 1, False, -1, True),
        (2, 1, False, 0, False),
        (3, 1, False, -1, True),
        (3, 1, False, 0, False),
    ],
)
def test_integer_out_of_bounds(
    total_bits: int, frac_bits: int, signed: bool, val_in: int, expected: bool
) -> None:
    result = FxpParams(
        total_bits=total_bits, frac_bits=frac_bits, signed=signed
    ).integer_out_of_bounds(val_in)
    assert result == expected


@pytest.mark.parametrize(
    "total_bits, frac_bits, signed, val_in, expected",
    [
        (2, 1, True, 1.0, True),
        (2, 1, True, 0.5, False),
        (3, 1, True, 2.0, True),
        (3, 1, True, 1.5, False),
        (2, 1, True, 1.0, True),
        (2, 1, True, 0.5, False),
        (3, 1, True, 2.0, True),
        (3, 1, True, 1.5, False),
        (2, 1, False, 2.0, True),
        (2, 1, False, 1.5, False),
        (3, 1, False, 4.0, True),
        (3, 1, False, 3.5, False),
    ],
)
def test_rational_upper_limit(
    total_bits: int, frac_bits: int, signed: bool, val_in: int, expected: bool
) -> None:
    result = FxpParams(
        total_bits=total_bits, frac_bits=frac_bits, signed=signed
    ).rational_out_overflow(val_in)
    assert result == expected


@pytest.mark.parametrize(
    "total_bits, frac_bits, signed, val_in, expected",
    [
        (2, 1, True, -1.5, True),
        (2, 1, True, -1.0, False),
        (3, 1, True, -2.5, True),
        (3, 1, True, -2.0, False),
        (2, 1, False, -0.5, True),
        (2, 1, False, 0.0, False),
        (3, 1, False, -0.5, True),
        (3, 1, False, 0.0, False),
    ],
)
def test_rational_downer_limit(
    total_bits: int, frac_bits: int, signed: bool, val_in: int, expected: bool
) -> None:
    result = FxpParams(
        total_bits=total_bits, frac_bits=frac_bits, signed=signed
    ).rational_out_underflow(val_in)
    assert result == expected


@pytest.mark.parametrize(
    "total_bits, frac_bits, signed, val_in, expected",
    [
        (2, 1, True, -1.5, True),
        (2, 1, True, -1.0, False),
        (3, 1, True, -2.5, True),
        (3, 1, True, -2.0, False),
        (2, 1, True, 1.0, True),
        (2, 1, True, 0.5, False),
        (3, 1, True, 2.0, True),
        (3, 1, True, 1.5, False),
        (2, 1, False, -0.5, True),
        (2, 1, False, 0.0, False),
        (3, 1, False, -0.5, True),
        (3, 1, False, 0.0, False),
    ],
)
def test_rational_out_of_bounds(
    total_bits: int, frac_bits: int, signed: bool, val_in: int, expected: bool
) -> None:
    result = FxpParams(
        total_bits=total_bits, frac_bits=frac_bits, signed=signed
    ).rational_out_of_bounds(val_in)
    assert result == expected



import pytest

from elasticai.creator.testing.cocotb_pytest import create_name_for_build_test_subdir


def simple_function():
    pass


def test_create_name_for_simple_function():
    actual = create_name_for_build_test_subdir(simple_function)
    expected = "simple_function"
    assert actual == expected


def function_with_args(a, b, c):
    pass


def test_create_name_function_with_args():
    actual = create_name_for_build_test_subdir(function_with_args, 1, 2, c="abcd")
    expected = "function_with_args_a_1_b_2_c_abcd"
    assert actual == expected


def test_create_name_for_function_with_list_args():
    actual = create_name_for_build_test_subdir(
        function_with_args, [1, 2, 3], 2, c="abcd"
    )
    h = hash((1, 2, 3)).to_bytes(length=8, signed=True).hex()
    expected = f"function_with_args_a_{h}_b_2_c_abcd"
    assert actual == expected


def test_create_name_for_fn_with_dict_args():
    actual = create_name_for_build_test_subdir(
        function_with_args, {"a": 1}, 2, c="abcd"
    )
    h = hash((("a", 1),)).to_bytes(length=8, signed=True).hex()
    expected = f"function_with_args_a_{h}_b_2_c_abcd"
    assert actual == expected


@pytest.fixture
def dummy_fixture_fn(request):
    def run(local_namespace):
        print(request.node.callspec)
        namespace_without_fixture = {
            k: v for k, v in local_namespace.items() if k != "dummy_fixture_fn"
        }
        return create_name_for_build_test_subdir(
            request.function, **namespace_without_fixture
        )

    return run


@pytest.mark.parametrize("x", [4])
def test_pass_parameter_to_fixture(dummy_fixture_fn, x):
    actual = dummy_fixture_fn(locals())
    expected = f"test_pass_parameter_to_fixture_x_{x}"
    assert actual == expected



from typing import TypeVar

from elasticai.creator.graph._types import NodeConstraintFn

from .graph import Graph
from .vf2 import find_all_matches, match

TP = TypeVar("TP")
T = TypeVar("T")


class SubGraphMatchError(Exception):
    def __init__(self):
        super().__init__("No matching subgraph found")


def find_subgraph(
    pattern: Graph[TP], graph: Graph[T], node_constraint: NodeConstraintFn[TP, T]
) -> dict[TP, T]:
    return match(pattern, graph, node_constraint)


def find_all_subgraphs(
    pattern: Graph[TP], graph: Graph[T], node_constraint: NodeConstraintFn[TP, T]
) -> list[dict[TP, T]]:
    return find_all_matches(pattern, graph, node_constraint)



from dataclasses import dataclass

import torch
from torch.nn.parameter import Parameter

from elasticai.creator.base_modules.lstm_cell import LSTMCell, MathOperations
from elasticai.creator.base_modules.math_operations_torch import TorchMathOperations
from tests.tensor_test_case import TensorTestCase


def tensor(x: list) -> torch.Tensor:
    return torch.tensor(x, dtype=torch.float32)


def create_lstm_cell_and_reference(
    input_size: int,
    hidden_size: int,
    bias: bool,
    operations: MathOperations = TorchMathOperations(),
) -> tuple[LSTMCell, torch.nn.LSTMCell]:
    cell = LSTMCell(
        input_size=input_size,
        hidden_size=hidden_size,
        bias=bias,
        sigmoid_factory=torch.nn.Sigmoid,
        tanh_factory=torch.nn.Tanh,
        operations=operations,
    )
    reference_cell = torch.nn.LSTMCell(
        input_size=input_size, hidden_size=hidden_size, bias=bias
    )

    def ones_like(x: torch.Tensor) -> torch.Tensor:
        return Parameter(torch.ones_like(x))

    cell.linear_ih.weight = ones_like(cell.linear_ih.weight)
    cell.linear_hh.weight = ones_like(cell.linear_hh.weight)
    reference_cell.weight_ih = ones_like(reference_cell.weight_ih)
    reference_cell.weight_hh = ones_like(reference_cell.weight_hh)
    if bias:
        cell.linear_ih.bias = ones_like(cell.linear_ih.bias)
        cell.linear_hh.bias = ones_like(cell.linear_hh.bias)
        reference_cell.bias_ih = ones_like(reference_cell.bias_ih)
        reference_cell.bias_hh = ones_like(reference_cell.bias_hh)
    return cell, reference_cell


@dataclass
class FixedMulOperations(TorchMathOperations):
    values: list[float]

    def mul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return tensor(self.values)


@dataclass
class FixedAddOperations(TorchMathOperations):
    values: list[float]

    def add(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return tensor(self.values)


class ZeroActivation(torch.nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.zeros_like(x)


class LSTMCellTest(TensorTestCase):
    def test_lstm_cell_without_bias_equals_torch_lstm_cell(self) -> None:
        cell, reference_cell = create_lstm_cell_and_reference(
            input_size=3, hidden_size=2, bias=False
        )
        inputs = tensor([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        actual_h, actual_c = cell(inputs)
        expected_h, expected_c = reference_cell(inputs)

        self.assertTensorEqual(expected_h, actual_h)
        self.assertTensorEqual(expected_c, actual_c)

    def test_lstm_cell_equals_torch_lstm_cell(self) -> None:
        cell, reference_cell = create_lstm_cell_and_reference(
            input_size=3, hidden_size=2, bias=True
        )
        inputs = tensor([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        actual_h, actual_c = cell(inputs)
        expected_h, expected_c = reference_cell(inputs)

        self.assertTensorEqual(expected_h, actual_h)
        self.assertTensorEqual(expected_c, actual_c)

    def test_lstm_cell_without_batched_inputs_equals_pytorch_lstm_cell(self) -> None:
        cell, reference_cell = create_lstm_cell_and_reference(
            input_size=4, hidden_size=2, bias=True
        )
        inputs = tensor([1, 2, 3, 4])
        actual_h, actual_c = cell(inputs)
        expected_h, expected_c = reference_cell(inputs)

        self.assertTensorEqual(expected_h, actual_h)
        self.assertTensorEqual(expected_c, actual_c)

    def test_lstm_cell_uses_mul_from_operations(self) -> None:
        cell = LSTMCell(
            input_size=1,
            hidden_size=2,
            bias=True,
            operations=FixedMulOperations([0.0]),
            sigmoid_factory=torch.nn.Sigmoid,
            tanh_factory=torch.nn.Tanh,
        )

        inputs = tensor([[1], [2]])
        h, c = cell(inputs)

        self.assertTensorEqual(h, [0.0])
        self.assertTensorEqual(c, [0.0])

    def test_lstm_cell_uses_add_from_operations(self) -> None:
        cell = LSTMCell(
            input_size=1,
            hidden_size=2,
            bias=False,
            operations=FixedAddOperations([0.0]),
            sigmoid_factory=torch.nn.Sigmoid,
            tanh_factory=torch.nn.Tanh,
        )

        inputs = tensor([[1], [2]])
        h, c = cell(inputs)

        self.assertTensorEqual(h, [0.0])
        self.assertTensorEqual(c, [0.0])

    def test_lstm_cell_uses_sigmoid_factory(self) -> None:
        cell = LSTMCell(
            input_size=1,
            hidden_size=2,
            bias=False,
            operations=TorchMathOperations(),
            sigmoid_factory=ZeroActivation,
            tanh_factory=torch.nn.Tanh,
        )

        inputs = tensor([[1]])
        h, c = cell(inputs)

        self.assertTensorEqual(h, [[0.0, 0.0]])
        self.assertTensorEqual(c, [[0.0, 0.0]])

    def test_lstm_cell_uses_tanh_factory(self) -> None:
        cell = LSTMCell(
            input_size=1,
            hidden_size=2,
            bias=False,
            operations=TorchMathOperations(),
            sigmoid_factory=torch.nn.Sigmoid,
            tanh_factory=ZeroActivation,
        )

        inputs = tensor([[1]])
        h, c = cell(inputs)

        self.assertTensorEqual(h, [[0.0, 0.0]])
        self.assertTensorEqual(c, [[0.0, 0.0]])



from pathlib import Path

from elasticai.creator.ir2vhdl.testing import run_vunit_vhdl_testbenches

deps = [
    "elasticai.creator_plugins.{name}".format(name=name)
    for name in [
        "striding_shift_register",
        "shift_register",
        "counter",
    ]
]


if __name__ == "__main__":
    run_vunit_vhdl_testbenches(deps, Path(__file__).parent)



from elasticai.creator import graph as gr
from elasticai.creator.graph import produces_dangling_edge


def build_graph(adjacency_list):
    g = gr.BaseGraph()
    for src in adjacency_list:
        for dst in adjacency_list[src]:
            g.add_edge(src, dst)
    return g


def test_catch_dangling_edge_d_b():
    g = build_graph(
        {
            "a": ["b"],
            "b": ["c"],
            "d": ["b"],
        }
    )
    match = {"0": "a", "1": "b", "2": "c"}
    lhs = {"i0": "0", "i1": "2"}
    assert produces_dangling_edge(g, match, lhs.values())


def test_catch_dangling_edge_b_d():
    g = build_graph(
        {
            "a": ["b"],
            "b": ["c", "d"],
        }
    )
    match = {"0": "a", "1": "b", "2": "c"}
    lhs = {"i0": "0", "i1": "2"}
    assert produces_dangling_edge(g, match, lhs.values())


def test_let_pass_if_successor_b_is_in_interface():
    g = build_graph(
        {
            "a": ["b"],
            "b": ["c"],
            "d": ["b"],
        }
    )
    match = {"0": "a", "1": "b", "2": "c"}
    lhs = {"i0": "0", "i1": "1"}
    assert not produces_dangling_edge(g, match, lhs.values())


def test_let_pass_if_predecessor_b_is_in_interface():
    g = build_graph(
        {
            "a": ["b"],
            "b": ["c", "d"],
        }
    )
    match = {"0": "a", "1": "b", "2": "c"}
    lhs = {"i0": "0", "i1": "1"}
    assert not produces_dangling_edge(g, match, lhs.values())


def test_remove_second_match_if_it_overlaps_with_first():
    g = build_graph(
        {
            "a": ["b"],
            "b": ["c"],
            "c": ["d"],
        }
    )
    first_match = {"0": "a", "1": "b", "2": "c"}
    second_match = {"0": "b", "1": "c", "2": "d"}
    lhs = {"i0": "0", "i1": "2"}
    assert [first_match] == list(
        gr.get_rewriteable_matches(
            g,
            [first_match, second_match],
            lhs.values(),
        )
    )


def test_do_not_remove_second_match_if_overlap_only_occurs_in_interface():
    g = build_graph(
        {
            "a": ["b"],
            "b": ["c"],
            "c": ["d"],
            "d": ["e"],
        }
    )
    first_match = {"0": "a", "1": "b", "2": "c"}
    second_match = {"0": "c", "1": "d", "2": "e"}
    lhs = {"i0": "0", "i1": "2"}
    assert [first_match, second_match] == list(
        gr.get_rewriteable_matches(
            g,
            [first_match, second_match],
            lhs.values(),
        )
    )


def test_pattern_seq_with_2_interface_nodes():
    g = build_graph(
        {
            "a": ["b"],
            "b": ["c"],
            "c": ["d"],
            "d": ["e"],
            "e": ["f"],
            "f": ["g"],
            "g": ["h"],
        }
    )
    p = build_graph(
        {
            "0": ["1"],
            "1": ["2"],
            "2": ["3"],
            "3": ["4"],
        }
    )
    matches = gr.find_all_subgraphs(
        pattern=p, graph=g, node_constraint=lambda _, __: True
    )
    lhs = {"i0": "0", "i1": "1", "i2": "3", "i3": "4"}
    expected_leftover_matches = []
    for sequence in [("a", "b", "c", "d", "e"), ("d", "e", "f", "g", "h")]:
        expected_leftover_matches.append(
            {f"{i}": node for i, node in enumerate(sequence)}
        )
    assert expected_leftover_matches == list(
        gr.get_rewriteable_matches(
            g,
            matches,
            lhs.values(),
        )
    )



import math

from elasticai.creator.file_generation.savable import Path
from elasticai.creator.file_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.vhdl.auto_wire_protocols.port_definitions import create_port
from elasticai.creator.vhdl.design.design import Design
from elasticai.creator.vhdl.design.ports import Port
from elasticai.creator.vhdl.shared_designs.rom import (
    Rom,
)

from .testbench import Conv1dDesignProtocol


def generate_parameters_from_port(port: Port) -> dict[str, str]:
    params = {}
    for signal in port:
        if signal.width > 0:
            params[f"{signal.name}_width"] = str(signal.width)
    return params


class Conv1dDesign(Design, Conv1dDesignProtocol):
    def __init__(
        self,
        name: str,
        total_bits: int,
        frac_bits: int,
        in_channels: int,
        out_channels: int,
        signal_length: int,
        kernel_size: int,
        weights: list[list[list[int]]],
        bias: list[int],
    ) -> None:
        super().__init__(name=name)
        self._total_bits = total_bits
        self._frac_bits = frac_bits
        self._in_channels = in_channels
        self._out_channels = out_channels
        self._input_signal_length = signal_length
        self._kernel_size = kernel_size
        self._weights = weights
        self._bias = bias
        self.output_signal_length = math.floor(
            self.input_signal_length - self.kernel_size + 1
        )
        self._port = create_port(
            x_width=self._total_bits,
            y_width=self._total_bits,
            x_count=self.input_signal_length * self._in_channels,
            y_count=self.output_signal_length * self._out_channels,
        )

    @property
    def input_signal_length(self) -> int:
        return self._input_signal_length

    @property
    def kernel_size(self) -> int:
        return self._kernel_size

    @property
    def port(self) -> Port:
        return self._port

    @property
    def in_channels(self) -> int:
        return self._in_channels

    @property
    def out_channels(self) -> int:
        return self._out_channels

    @staticmethod
    def _flatten_params(params: list[list[list[int]]]) -> list[int]:
        result = []
        for list_of_lists in params:
            for list_of_int in list_of_lists:
                result.extend(list_of_int)
        return result

    def save_to(self, destination: Path) -> None:
        print(self.name)
        rom_name = dict(weights=f"{self.name}_w_rom", bias=f"{self.name}_b_rom")
        template = InProjectTemplate(
            package=module_to_package(self.__module__),
            file_name="conv1d.tpl.vhd",
            parameters=dict(
                frac_width=str(self._frac_bits),
                in_channels=str(self._in_channels),
                out_channels=str(self._out_channels),
                kernel_size=str(self.kernel_size),
                vector_width=str(self.input_signal_length),
                name=self.name,
            )
            | generate_parameters_from_port(self._port),
        )
        destination.create_subpath(self.name).as_file(".vhd").write(template)

        core_component = InProjectTemplate(
            package=module_to_package(self.__module__),
            file_name="conv1d_function.tpl.vhd",
            parameters={
                "rom_name_weights": rom_name["weights"],
                "rom_name_bias": rom_name["bias"],
            },
        )
        destination.create_subpath(f"{self.name}_fxp_MAC_RoundToZero").as_file(
            ".vhd"
        ).write(core_component)

        mac = InProjectTemplate(
            package="elasticai.creator.nn.fixed_point.mac",
            file_name="fxp_mac.tpl.vhd",
            parameters={},
        )
        destination.create_subpath("fxp_mac").as_file(".vhd").write(mac)

        weights_rom = Rom(
            name=rom_name["weights"],
            data_width=self._total_bits,
            values_as_integers=self._flatten_params(self._weights),
        )
        weights_rom.save_to(destination)

        bias_rom = Rom(
            name=rom_name["bias"],
            data_width=self._total_bits,
            values_as_integers=self._bias,
        )
        bias_rom.save_to(destination)



from elasticai.creator.ir2vhdl import (
    Edge,
    edge,
    vhdl_node,
)
from elasticai.creator.ir2vhdl import (
    Implementation as _Implementation,
)
from elasticai.creator_plugins.grouped_filter import (
    FilterParameters,
)
from elasticai.creator_plugins.grouped_filter import (
    grouped_filter as filter,
)


class Implementation(_Implementation):
    def input(self, shape: tuple[int, int]):
        self.add_node(
            vhdl_node(
                name="input",
                type="input",
                implementation="",
                input_shape=shape,
                output_shape=shape,
            )
        )

    def output(self, shape: tuple[int, int]):
        self.add_node(
            vhdl_node(
                name="output",
                type="output",
                implementation="",
                output_shape=shape,
                input_shape=shape,
            )
        )


def high_level_ir(groups: int) -> Implementation:
    p = FilterParameters(
        kernel_size=2, in_channels=2, out_channels=4, input_size=10, groups=groups
    )
    hl_ir = Implementation(
        name="filter",
        type="lutron_filter",
        data={
            "filter_parameters": p.as_dict(),
            "kernel_per_group": tuple(f"lutron_{i}" for i in range(groups)),
        },
    )
    hl_ir.input((p.in_channels, p.input_size))
    hl_ir.output((p.out_channels, p.num_steps))
    hl_ir.add_node(
        vhdl_node(
            name="lutron_0",
            input_shape=(p.in_channels * p.kernel_size,),
            output_shape=(p.out_channels,),
            type="lutron",
            implementation="lutron_0",
        )
    )
    return hl_ir


class TestLutronFilterWithSingleGroup:
    def test_has_correct_edges(self) -> None:
        hl_ir = high_level_ir(1)
        p = FilterParameters.from_dict(hl_ir.attributes["filter_parameters"])
        expected_edges: set[Edge] = set()
        expected_edges.add(
            edge(
                src="input",
                dst="lutron_0_i0",
                src_dst_indices=tuple(
                    zip(
                        range(p.in_channels * p.kernel_size),
                        range(p.in_channels * p.kernel_size),
                    )
                ),
            )
        )
        expected_edges.add(
            edge(
                src="lutron_0_i0",
                dst="output",
                src_dst_indices=(
                    f"range(0, {p.out_channels})",
                    f"range(0, {p.out_channels})",
                ),
            )
        )
        actual_ir = filter(hl_ir)
        assert set(actual_ir.edges.values()) == expected_edges


class TestLutronFilterWithTwoGroups:
    def test_has_correct_edges(self) -> None:
        hl_ir = high_level_ir(2)
        FilterParameters.from_dict(hl_ir.attributes["filter_parameters"])
        expected_edges: set[Edge] = set()
        expected_edges = expected_edges.union(
            (
                edge(
                    src="input",
                    dst="lutron_0_i0",
                    src_dst_indices=(
                        (0, 0),
                        (2, 1),
                    ),
                ),
                edge(
                    src="input",
                    dst="lutron_1_i0",
                    src_dst_indices=((1, 0), (3, 1)),
                ),
                edge(
                    src="lutron_0_i0",
                    dst="output",
                    src_dst_indices=("range(0, 2)", "range(0, 2)"),
                ),
                edge(
                    src="lutron_1_i0",
                    dst="output",
                    src_dst_indices=("range(0, 2)", "range(2, 4)"),
                ),
            )
        )
        actual_ir = filter(hl_ir)
        assert set(actual_ir.edges.values()) == expected_edges



from typing import cast

from torch import nn

from elasticai.creator.torch2ir import Implementation

handlers = []


def _register(fn):
    handlers.append(fn)
    return fn


@_register
def linear(impl: Implementation) -> nn.Module:
    return nn.Linear(
        in_features=cast(int, impl.data["in_features"]),
        out_features=cast(int, impl.data["out_features"]),
        bias=cast(bool, impl.data["bias"]),
    )


@_register
def relu(impl: Implementation) -> nn.Module:
    return nn.ReLU()


@_register
def conv1d(impl: Implementation) -> nn.Conv1d:
    keywords = (
        "in_channels",
        "out_channels",
        "bias",
        "groups",
        "kernel_size",
        "stride",
    )
    kwargs = dict(((k, impl.data[k]) for k in keywords))
    return nn.Conv1d(**kwargs)  # type: ignore


@_register
def sigmoid(impl: Implementation) -> nn.Sigmoid:
    return nn.Sigmoid()


@_register
def flatten(imp: Implementation) -> nn.Flatten:
    return nn.Flatten()



class FilterParameters:
    """
    Represents all necessary parameters to implement a (grouped) 1d filter
    without padding.
    args:
        input_size: Denotes the spatial input size of the filter, consequently we assume `kernel_size <= input_size`.
            However, in some cases `input_size` might not be known, until this changes we set it to zero.
            Typically, we need to run an inference once to set the input_size fields.
        kernel_size: Either the size of the kernel in case of e.g. conv1d or pooling layers, or `-1` in case of a layer
            that expects a flat tensor (e.g., linear layer).
            IMPORTANT: The way we represent linear layers is expected to change in the future.
        input_channels: The number of input channels in case of a 1d filter. In case of a flat input tensor, e.g., linear layer
    """

    _field_names = (
        "out_channels",
        "in_channels",
        "kernel_size",
        "groups",
        "stride",
        "input_size",
        "output_size",
    )

    def __init__(
        self,
        kernel_size: int,
        in_channels: int,
        out_channels: int,
        groups: int = 1,
        stride: int = 1,
        input_size: int | None = None,
        output_size: int = 1,
    ):
        self._in_channels = in_channels
        self._out_channels = out_channels
        self._stride = stride
        self._groups = groups
        self.kernel_size = kernel_size
        self._check_group_validity()
        if input_size is None:
            input_size = kernel_size
        self.input_size = input_size
        self.output_size = output_size

    def _handle_kernel_size(self, kernel_size: int | tuple[int, ...]) -> int:
        if isinstance(kernel_size, tuple):
            if len(kernel_size) > 1:
                raise ValueError("unsupported 2d kernel, only 1d kernels are supported")
            kernel_size = kernel_size[0]
        return kernel_size

    @property
    def kernel_size(self) -> int:
        return self._kernel_size

    @kernel_size.setter
    def kernel_size(self, kernel_size: int | tuple[int]) -> None:
        self._kernel_size = self._handle_kernel_size(kernel_size)

    def _check_group_validity(self):
        if not (
            self.in_channels % self.groups == 0 and self.out_channels % self.groups == 0
        ):
            raise ValueError(
                f"groups has to be a divider of in_channels and out_channels, but found in_channels: {self.in_channels}, out_channels: {self.out_channels}, groups: {self.groups}"
            )

    @property
    def fan_in(self) -> int:
        return self.in_channels // self.groups * self.kernel_size

    @property
    def stride(self) -> int:
        return self._stride

    @property
    def num_steps(self) -> int:
        return (self.input_size - self.kernel_size) // self.stride + 1

    @property
    def groups(self) -> int:
        return self._groups

    @groups.setter
    def groups(self, groups: int):
        self._groups = groups
        self._check_group_validity()

    @property
    def in_channels(self) -> int:
        return self._in_channels

    @in_channels.setter
    def in_channels(self, in_channels: int):
        self._in_channels = in_channels
        self._check_group_validity()

    @property
    def in_channels_per_group(self) -> int:
        return self.in_channels // self.groups

    @property
    def out_channels_per_group(self) -> int:
        return self.out_channels // self.groups

    @property
    def out_channels(self) -> int:
        return self._out_channels

    @out_channels.setter
    def out_channels(self, out_channels: int):
        self._out_channels = out_channels
        self._check_group_validity()

    @property
    def total_skipped_inputs_per_step(self) -> int:
        return self.stride * self.in_channels

    def get_in_channels_by_group(self) -> tuple[tuple[int, ...], ...]:
        return self._get_channels_by_group(self.in_channels_per_group)

    def get_out_channels_by_group(self) -> tuple[tuple[int, ...], ...]:
        return self._get_channels_by_group(self.out_channels_per_group)

    def _get_channels_by_group(self, channels_per_group):
        def get_channels_for_group(g):
            return tuple(range(g * channels_per_group, (g + 1) * channels_per_group))

        return tuple((get_channels_for_group(g) for g in range(self.groups)))

    def with_groups(self, groups) -> "FilterParameters":
        return self.__class__(
            kernel_size=self.kernel_size,
            groups=groups,
            out_channels=self.out_channels,
            in_channels=self.in_channels,
        )

    def _value_dict(self):
        return {f: getattr(self, f) for f in self._field_names}

    def as_dict(self):
        return self._value_dict()

    @classmethod
    def from_dict(cls, d: dict):
        return cls(**{k: d[k] for k in cls._field_names})

    def __repr__(self) -> str:
        params = ", ".join([f"{k}={v}" for k, v in self._value_dict().items()])
        return f"FilterParameters({params})"

    def __hash__(self):
        return hash(tuple(self._value_dict().values()))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        else:
            return hash(self) == hash(other)



from typing import cast

import pytest

from elasticai.creator.file_generation.in_memory_path import InMemoryFile, InMemoryPath
from elasticai.creator.nn.fixed_point.conv1d.design import Conv1dDesign


@pytest.fixture
def conv1d_design() -> Conv1dDesign:
    return Conv1dDesign(
        name="conv1d",
        total_bits=16,
        frac_bits=8,
        in_channels=1,
        out_channels=2,
        kernel_size=3,
        signal_length=4,
        weights=[[[1, 1, 1]], [[1, 1, 1]]],
        bias=[1, 1],
    )


def save_design(design: Conv1dDesign) -> dict[str, str]:
    destination = InMemoryPath("conv1d", parent=None)
    design.save_to(destination)
    files = cast(list[InMemoryFile], list(destination.children.values()))
    return {file.name: "\n".join(file.text) for file in files}


def test_saved_design_contains_needed_files(conv1d_design: Conv1dDesign) -> None:
    saved_files = save_design(conv1d_design)

    expected_files = {
        "conv1d_w_rom.vhd",
        "conv1d_b_rom.vhd",
        "conv1d.vhd",
        "conv1d_fxp_MAC_RoundToZero.vhd",
        "fxp_mac.vhd",
    }
    actual_files = set(saved_files.keys())

    assert expected_files == actual_files


def test_weight_rom_code_generated_correctly(conv1d_design: Conv1dDesign) -> None:
    expected_code = """library ieee;
use ieee.numeric_std.all;
use ieee.std_logic_1164.all;

entity conv1d_w_rom is
    generic (
        ROM_ADDR_WIDTH : integer := 3;
        ROM_DATA_WIDTH : integer := 16
    );
    port (
        clk : in std_logic;
        en : in std_logic;
        addr : in std_logic_vector(ROM_ADDR_WIDTH-1 downto 0);
        data : out std_logic_vector(ROM_DATA_WIDTH-1 downto 0)
    );
end entity conv1d_w_rom;
architecture rtl of conv1d_w_rom is
    type conv1d_w_rom_array_t is array (0 to 2**ROM_ADDR_WIDTH-1) of std_logic_vector(ROM_DATA_WIDTH-1 downto 0);
    signal ROM : conv1d_w_rom_array_t:=("0000000000000001", "0000000000000001", "0000000000000001", "0000000000000001", "0000000000000001", "0000000000000001", "0000000000000000", "0000000000000000");
    attribute rom_style : string;
    attribute rom_style of ROM : signal is "auto";
begin
    ROM_process: process(addr)
    begin
        if (en = '1') then
            data <= ROM(to_integer(unsigned(addr)));
        end if;
    end process ROM_process;
end architecture rtl;"""
    saved_files = save_design(conv1d_design)
    actual_code = saved_files["conv1d_w_rom.vhd"]
    assert expected_code == actual_code


def test_bias_rom_code_generated_correctly(conv1d_design: Conv1dDesign) -> None:
    expected_code = """library ieee;
use ieee.numeric_std.all;
use ieee.std_logic_1164.all;

entity conv1d_b_rom is
    generic (
        ROM_ADDR_WIDTH : integer := 1;
        ROM_DATA_WIDTH : integer := 16
    );
    port (
        clk : in std_logic;
        en : in std_logic;
        addr : in std_logic_vector(ROM_ADDR_WIDTH-1 downto 0);
        data : out std_logic_vector(ROM_DATA_WIDTH-1 downto 0)
    );
end entity conv1d_b_rom;
architecture rtl of conv1d_b_rom is
    type conv1d_b_rom_array_t is array (0 to 2**ROM_ADDR_WIDTH-1) of std_logic_vector(ROM_DATA_WIDTH-1 downto 0);
    signal ROM : conv1d_b_rom_array_t:=("0000000000000001", "0000000000000001");
    attribute rom_style : string;
    attribute rom_style of ROM : signal is "auto";
begin
    ROM_process: process(addr)
    begin
        if (en = '1') then
            data <= ROM(to_integer(unsigned(addr)));
        end if;
    end process ROM_process;
end architecture rtl;"""
    saved_files = save_design(conv1d_design)
    actual_code = saved_files["conv1d_b_rom.vhd"]
    print(actual_code)
    assert expected_code == actual_code



import torch

from elasticai.creator.arithmetic import FxpArithmetic, FxpParams
from tests.system_tests.helper.parse_tensors_to_bytearray import (
    parse_bytearray_to_fxp_tensor,
    parse_fxp_tensor_to_bytearray,
)


def test_tensors_to_bytearray():
    torch.manual_seed(0)
    total_bits = 8
    frac_bits = 2
    fxp_conf = FxpArithmetic(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    )
    tensor = fxp_conf.as_rational(fxp_conf.cut_as_integer(torch.randn(2, 3, 4)))
    result = parse_fxp_tensor_to_bytearray(tensor, total_bits, frac_bits)
    expected = [
        bytearray(b"\xfc\x03\x01\xfc\x02\x00\xff\xff\x05\xff\xf8\x01"),
        bytearray(b"\x00\x03\x02\x03\x05\xff\x00\x08\xfd\x00\x00\xf7"),
    ]
    assert result == expected


def test_bytearray_to_tensor():
    input = [
        bytearray(b"\xfc\x03\x01\xfc\x02\x00\xff\xff\x05\xff\xf8\x01"),
        bytearray(b"\x00\x03\x02\x03\x05\xff\x00\x08\xfd\x00\x00\xf7"),
    ]
    total_bits = 8
    frac_bits = 2
    torch.manual_seed(0)
    fxp_conf = FxpArithmetic(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    )
    dimensions = (2, 3, 4)
    expected = fxp_conf.as_rational(fxp_conf.cut_as_integer(torch.randn(dimensions)))
    result = parse_bytearray_to_fxp_tensor(input, total_bits, frac_bits, dimensions)
    assert torch.equal(result, expected)



import cocotb
from cocotb.triggers import Timer


def model(a: int, scale: int, offset: int, do_offset: bool) -> int:
    return scale * a + (0 if not do_offset else offset)


@cocotb.test()
async def model_test(dut):
    A = 1
    B = dut.OFFSET.value.to_signed() if "OFFSET" in dir(dut) else 0
    dut.A.value = A
    await Timer(2, unit="step")
    assert dut.Q.value == model(A, dut.SCALE.value.to_signed(), B, "OFFSET" in dir(dut))



from typing import Callable

import torch
from torch import Tensor, optim
from torch.nn import Module

from elasticai.creator_plugins.quantized_grads.base_modules import parametrized_modules


class _QOptim(optim.Optimizer):
    def __init__(
        self,
        model,
        *args,
        buffer_quantizations: dict[str, Callable[[Tensor], Tensor]] = None,
        **kwargs,
    ): ...


def get_quantized_optimizer(optimizer: type[optim.Optimizer]) -> type[_QOptim]:
    class QuantizedOptim(optimizer):
        def __init__(
            self,
            model: Module,
            *args,
            buffer_quantizations: dict[str, Module] = None,
            **kwargs,
        ):
            super().__init__(*args, **kwargs)
            self._model = model
            self.register_step_post_hook(_ensure_tensors_quantization_after_update)
            if buffer_quantizations is None:
                self.buffer_quantization = {}
            else:
                self.buffer_quantization: dict[str, Callable[[Tensor], Tensor]] = (
                    buffer_quantizations
                )
                self.register_step_post_hook(_quantization_for_buffers_in_optimizer)

        def step(self, closure=None):
            super().step(closure)

    def _ensure_tensors_quantization_after_update(
        optimizer: QuantizedOptim, *args, **kwargs
    ):
        with torch.no_grad():
            for m in optimizer._model.modules():
                if m.__class__.__name__ in parametrized_modules:
                    for p in m.parametrizations:
                        m_parametrization_param = m.parametrizations[p]
                        m_param = getattr(m, p)
                        m_param.data = m_parametrization_param[0].right_inverse(m_param)

    def _quantization_for_buffers_in_optimizer(
        optimizer: QuantizedOptim, *args, **kwargs
    ):
        for group in optimizer.param_groups:
            for p in group["params"]:
                state = optimizer.state[p]
                for key in state.keys():
                    if key in optimizer.buffer_quantization.keys():
                        if (
                            state[key].device
                            != optimizer.buffer_quantization[key].device
                        ):
                            optimizer.buffer_quantization[key].to(state[key].device)
                        state[key] = optimizer.buffer_quantization[key](state[key])

    return QuantizedOptim



import subprocess
import time
import tomllib
from pathlib import Path

import numpy as np
import pytest
import serial  # type: ignore
import torch
from elasticai.runtime.env5.usb import UserRemoteControl, get_env5_port  # type: ignore

from elasticai.creator.arithmetic import (
    FxpArithmetic,
    FxpParams,
)
from elasticai.creator.file_generation.on_disk_path import OnDiskPath
from elasticai.creator.nn import Sequential
from elasticai.creator.nn.fixed_point.linear import Linear
from elasticai.creator.vhdl.system_integrations.firmware_env5 import FirmwareENv5
from tests.system_tests.helper.parse_tensors_to_bytearray import (
    parse_bytearray_to_fxp_tensor,
    parse_fxp_tensor_to_bytearray,
)


def create_vhd_files(
    output_dir: str,
    num_inputs: int,
    num_outputs: int,
    total_bits: int,
    frac_bits: int,
    skeleton_id: list[int],
) -> Sequential:
    nn = Sequential(
        Linear(
            in_features=num_inputs,
            out_features=num_outputs,
            bias=True,
            total_bits=total_bits,
            frac_bits=frac_bits,
        )
    )
    nn[0].weight.data = torch.ones_like(nn[0].weight) * 2
    nn[0].bias.data = torch.Tensor([1.0, 2.0, -1.0])
    destination = OnDiskPath(output_dir)
    my_design = nn.create_design("nn")
    my_design.save_to(destination.create_subpath("srcs"))

    firmware = FirmwareENv5(
        network=my_design,
        x_num_values=num_inputs,
        y_num_values=num_outputs,
        id=skeleton_id,
        skeleton_version="v2",
    )
    firmware.save_to(destination)
    return nn


def vivado_build_binfile(input_dir: str, binfile_dir: str) -> None:
    print(f"Building binfile in {binfile_dir}")
    time.sleep(5)
    with open("./tests/system_tests/vivado_build_server_conf.toml", "rb") as f:
        config = tomllib.load(f)
    subprocess.run(
        [
            "bash",
            "./utils/autobuild_binfile_vivado2021.1.sh",
            config["username"],
            config["ip"],
            input_dir,
            binfile_dir,
        ],
        capture_output=True,
    )
    # print(f"{out.stdout=}")


@pytest.mark.hardware
def test_linear_layer_env5():
    # --- Settings
    # dev_address = "COM8"
    dev_address = get_env5_port()
    vivado_build = True

    # torch.manual_seed(1)
    total_bits = 8
    frac_bits = 2
    num_inputs = 5
    num_outputs = 3
    skeleton_id = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]

    # --- Processing
    # --- Creating the dummy
    input_tensor = torch.Tensor([[[0.0, 0.0, 0.0, 0.0, 0.0]]])
    for idx_array in range(0, num_inputs):
        for value in np.arange(-2, 2, 0.5):
            list_zeros = [0.0 for idx in range(0, num_inputs)]
            list_zeros[idx_array] = value
            input_tensor = torch.cat(
                (input_tensor, torch.Tensor([[list_zeros]])), dim=0
            )

    skeleton_id_as_bytearray = bytearray()
    for x in skeleton_id:
        skeleton_id_as_bytearray.extend(
            x.to_bytes(length=1, byteorder="little", signed=False)
        )

    # --- Preparing the VHDL code generation
    vhdl_dir = "./tests/system_tests/linear_layer/build_dir"
    binfile_dir = "./tests/system_tests/linear_layer/build_dir_output"
    binfile_path = Path(binfile_dir + "/output/env5_top_reconfig.bin")
    nn = create_vhd_files(
        vhdl_dir, num_inputs, num_outputs, total_bits, frac_bits, skeleton_id
    )

    fxp_params = FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    fxp_conf = FxpArithmetic(fxp_params)
    inputs = fxp_conf.as_rational(fxp_conf.cut_as_integer(input_tensor))
    expected_outputs = nn(inputs)

    # --- Building the code
    if vivado_build:
        vivado_build_binfile(vhdl_dir, binfile_dir)

    # --- Open Serial Communication to Device
    with serial.Serial(dev_address) as serial_con:
        flash_start_address = 0
        urc = UserRemoteControl(device=serial_con)
        urc.send_and_deploy_model(
            binfile_path, flash_start_address, skeleton_id_as_bytearray
        )
        # urc.deploy_model(flash_start_address, skeleton_id_as_bytearray)

        # --- Doing the test
        batch_data = parse_fxp_tensor_to_bytearray(inputs, total_bits, frac_bits)
        inference_result = list()
        state = False
        urc.fpga_leds(True, False, False, False)
        for i, sample in enumerate(batch_data):
            urc.fpga_leds(True, False, False, state)
            state = False if state else True

            batch_result = urc.inference_with_data(sample, num_outputs)
            my_result = parse_bytearray_to_fxp_tensor(
                [batch_result], total_bits, frac_bits, (1, 1, 3)
            )

            dev_inp = my_result
            dev_out = expected_outputs.data[i].view((1, 1, 3))
            if not torch.equal(dev_inp, dev_out):
                print(
                    f"Batch #{i:02d}: \t{dev_inp} == {dev_out}, (Delta ="
                    f" {dev_inp - dev_out}) \t\t\t\tfor input {inputs[i]}"
                )
                if i % 4 == 3:
                    print("\n")

            inference_result.append(batch_result)

        urc.fpga_leds(False, False, False, False)
        actual_result = parse_bytearray_to_fxp_tensor(
            inference_result, total_bits, frac_bits, expected_outputs.shape
        )

        assert torch.equal(actual_result, expected_outputs)



from dataclasses import dataclass

import torch


@dataclass(frozen=True)
class FixedPointConfigV2:
    """
    This class behaves almost like a frozen instance,
    but calculates and sets minimum_as_rational_tensor and maximum_as_rational_tensor just once in __post__init__.
    After the __post_init__ the __setattr__ and __getattr__ methods block the change of variables.
    """

    total_bits: int
    frac_bits: int

    def __post_init__(self):
        """
        This python API call for dataclass is executed after init.
        This way we check if the FixedPointConfigV2 is valid while keeping the config immutable.
        """
        if self.total_bits <= 0:
            raise Exception(
                f"total bits need to be > 0 for {self.__class__.__name__}. "
                f"You have set {self.total_bits=}."
            )
        if self.frac_bits + 1 > self.total_bits:
            raise Exception(
                f"total bits-1 needs to be > frac bits for {self.__class__.__name__}."
                f"You have set {self.total_bits=} and {self.frac_bits=}."
            )

    @property
    def minimum_as_rational(self):
        return torch.Tensor([-(2 ** (self.total_bits - self.frac_bits - 1))])

    @property
    def maximum_as_rational(self):
        return torch.Tensor([-self.minimum_as_rational - 1 / (2**self.frac_bits)])

    @property
    def resolution_per_int(self):
        return torch.Tensor([2**self.frac_bits])



import hypothesis.strategies as st
import pytest
from hypothesis import settings as hypothesis_settings
from hypothesis.stateful import (
    RuleBasedStateMachine,
    initialize,
    invariant,
    precondition,
    rule,
)
from torch.nn import Conv1d, Flatten, Linear, ReLU, Sequential, Sigmoid

from elasticai.creator.ir2torch import get_default_converter as ir2torch
from elasticai.creator.torch2ir import get_default_converter as torch2ir


def model():
    return Sequential(
        Linear(1, 2),
        ReLU(),
    )


def convert(model):
    return torch2ir().convert(model)


@st.composite
def conv_channel_group_parameters(draw: st.DrawFn) -> tuple[int, int, int]:
    test_class_instance = draw(st.runner())
    out_channels = draw(st.integers(min_value=1, max_value=64))
    in_channels = test_class_instance.last_out_channels
    if in_channels == -1:
        in_channels = draw(st.integers(max_value=64))
    groups = draw(
        st.integers(min_value=1, max_value=out_channels).filter(
            lambda n: out_channels % n == 0 and in_channels % n == 0
        )
    )
    return in_channels, out_channels, groups


@hypothesis_settings(max_examples=200, stateful_step_count=20)
class BuildModelFromIr(RuleBasedStateMachine):
    @initialize(start_channels=st.integers(min_value=1, max_value=64))
    def setup(self, start_channels):
        self.model = Sequential()
        self.last_out_channels = start_channels
        self.is_flattened = False

    @rule(
        channels_and_groups=conv_channel_group_parameters(),
        bias=st.booleans(),
        stride=st.integers(min_value=1, max_value=4),
        kernel_size=st.integers(min_value=1, max_value=12),
        padding=st.integers(min_value=0, max_value=3),
    )
    @precondition(lambda self: not self.is_flattened)
    def append_conv(self, channels_and_groups, bias, stride, kernel_size, padding):
        in_channels, out_channels, groups = channels_and_groups
        self.model.append(
            Conv1d(
                in_channels=in_channels,
                out_channels=out_channels,
                stride=stride,
                groups=groups,
                kernel_size=kernel_size,
                padding=padding,
            )
        )
        self.last_out_channels = out_channels

    @rule()
    def append_relu(self):
        self.model.append(ReLU())

    @rule()
    def append_sigmoid(self):
        self.model.append(Sigmoid())

    @rule()
    @precondition(lambda self: not self.is_flattened)
    def flatten(self):
        self.model.append(Flatten())
        self.is_flattened = True

    @rule(out_features=st.integers(min_value=1, max_value=32), bias=st.booleans())
    @precondition(lambda self: self.is_flattened)
    def append_linear(self, out_features, bias):
        self.model.append(
            Linear(
                in_features=self.last_out_channels, out_features=out_features, bias=bias
            )
        )

    @invariant()
    def model_from_ir_and_state_dict_is_the_same(self):
        original = self.model
        ir = convert(original)
        state = original.state_dict()
        rebuilt = ir2torch().convert(ir, state)

        def to_native_python(data) -> dict:
            result = {}
            for name, p in data:
                result[name] = p.tolist() if hasattr(p, "tolist") else p
            return result

        original_params = to_native_python(original.named_parameters())
        rebuilt_params = to_native_python(rebuilt.named_parameters())
        assert original_params == rebuilt_params
        del rebuilt


BuildModelFromIrTest = pytest.mark.slow()(BuildModelFromIr.TestCase)



from collections import OrderedDict

import hypothesis.strategies as st
import torch
from hypothesis import given
from torch.nn import Linear, ReLU, Sequential

from elasticai.creator.ir2torch import get_default_converter as ir2torch_converter
from elasticai.creator.torch2ir import get_default_converter as torch2ir_converter


def model():
    return Sequential(
        Linear(1, 2),
        ReLU(),
    )


def convert(model):
    return torch2ir_converter().convert(model)


def to_native_python(data) -> dict:
    result = {}
    for name, p in data:
        result[name] = p.tolist() if hasattr(p, "tolist") else p
    return result


@given(st.tuples(st.integers(1, 10), st.integers(1, 10)), st.booleans())
def test_build_model_from_ir_and_state_dict(num_features, bias):
    in_features, out_features = num_features
    original = Sequential(Linear(in_features, out_features, bias))
    ir = convert(original)
    state = original.state_dict()
    rebuilt = ir2torch_converter().convert(ir, state)
    rebuilt.load_state_dict(state)

    def to_native_python(data) -> dict:
        result = {}
        for name, p in data:
            result[name] = p.tolist() if hasattr(p, "tolist") else p
        return result

    original_params = to_native_python(original.named_parameters())
    rebuilt_params = to_native_python(rebuilt.named_parameters())
    assert original_params == rebuilt_params


def test_can_rebuild_model_with_duplicate_submodule():
    lin = Linear(1, 1)
    original = Sequential(lin, lin)
    state = original.state_dict()
    ir = convert(original)
    rebuilt = ir2torch_converter().convert(ir, state)
    original_params = to_native_python(original.named_parameters())
    rebuilt_params = to_native_python(rebuilt.named_parameters())
    assert original_params == rebuilt_params


def test_create_parent_modules_for_implemantations_with_dots():
    original = Sequential(OrderedDict(top=Sequential(OrderedDict(nested=Linear(1, 1)))))
    ir = convert(original)
    rebuilt = ir2torch_converter().convert(ir, original.state_dict())
    original_params = to_native_python(original.named_parameters())
    rebuilt_params = to_native_python(rebuilt.named_parameters())
    assert original_params == rebuilt_params


def test_perform_inference():
    original = model()
    ir = convert(original)
    rebuilt = ir2torch_converter().convert(ir)
    print(rebuilt.graph)
    rebuilt(torch.randn((1, 1)))



from .move_to_submodules import (
    PatternNode,
    build_sequential_pattern,
    move_pattern_to_subimpls,
)
from .reorder import NodeConstraint, SequenceReorderer

__all__ = [
    "SequenceReorderer",
    "NodeConstraint",
    "move_pattern_to_subimpls",
    "build_sequential_pattern",
    "PatternNode",
]



from typing import cast

from elasticai.creator.file_generation.in_memory_path import InMemoryFile, InMemoryPath
from elasticai.creator.nn.fixed_point.precomputed.adaptable_silu import AdaptableSiLU


def test_vhdl_code_matches_expected_adapt_silu() -> None:
    expected = """library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity adapt_silu is
    generic (
        BITWIDTH_INPUT : integer := 8;
        BITWIDTH_OUTPUT : integer := 8
    );
    port (
        enable : in std_logic;
        clock  : in std_logic;
        x      : in std_logic_vector(BITWIDTH_INPUT-1 downto 0);
        y      : out std_logic_vector(BITWIDTH_OUTPUT-1 downto 0)
    );
end adapt_silu;

architecture rtl of adapt_silu is
    signal signed_x : signed(BITWIDTH_INPUT-1 downto 0) := (others=>'0');
    signal signed_y : signed(BITWIDTH_OUTPUT-1 downto 0) := (others=>'0');
begin
    signed_x <= signed(x);
    y <= std_logic_vector(signed_y);

    adapt_silu_process : process(signed_x)
    begin
        if enable = '0' then
            signed_y <= to_signed(0, BITWIDTH_OUTPUT);
        else
            if signed_x <= to_signed(-128, BITWIDTH_OUTPUT) then signed_y <= to_signed(-11, BITWIDTH_OUTPUT);
            elsif signed_x <= to_signed(-43, BITWIDTH_OUTPUT) then signed_y <= to_signed(-18, BITWIDTH_OUTPUT);
            elsif signed_x <= to_signed(42, BITWIDTH_OUTPUT) then signed_y <= to_signed(0, BITWIDTH_OUTPUT);
            else signed_y <= to_signed(67, BITWIDTH_OUTPUT);
            end if;
        end if;
    end process;
end rtl;
""".splitlines()
    sigmoid = AdaptableSiLU(
        total_bits=8,
        frac_bits=6,
        num_steps=4,
        sampling_intervall=(-float("inf"), float("inf")),
    )
    build_path = InMemoryPath("build", parent=None)
    design = sigmoid.create_design("adapt_silu")
    design.save_to(build_path)
    actual = cast(InMemoryFile, build_path["adapt_silu"]).text

    for text in actual:
        print(text)
    assert actual == expected



import importlib
import inspect
import json
from collections.abc import Callable, Iterable, Iterator, Mapping, Sequence
from contextlib import ExitStack
from functools import wraps
from os import environ
from pathlib import Path
from typing import Any

import pytest

from .cocotb_prepare import build_report_folder_and_testdata
from .cocotb_runner import run_cocotb_sim


def _get_args(fn: Callable, *args: Any, **kwargs: Any) -> dict[str, Any]:
    return inspect.signature(fn).bind_partial(*args, **kwargs).arguments


def create_name_for_build_test_subdir(fn: Callable, *args: Any, **kwargs: Any) -> str:
    mangling_args: list[str] = []

    def get_hash(v):
        return hash(v).to_bytes(length=8, signed=True).hex()

    for arg, value in _get_args(fn, *args, **kwargs).items():
        if isinstance(value, str):
            pass
        elif isinstance(value, Sequence):
            value = get_hash(tuple(value))
        elif isinstance(value, Mapping):
            value = get_hash(tuple((k, v) for k, v in value.items()))
        mangling_args.extend([str(arg), str(value)])
    if len(mangling_args) > 0:
        argstring = "_" + "_".join(mangling_args)
    else:
        argstring = ""
    return f"{fn.__name__}{argstring}"


def eai_testbench(fn):
    """
    Intended usage:

    ```python
    @cocotb.test()
    @eai_testbench
    async def my_testbench_for_input_buffer(dut, x, input_data):
      dut.d_in = x
    ```

    and

    ```python
    @pytest.mark.parametrize("x", [1, 2, 3])
    def test_input_buffer(cocotb_test_fixture, x):
      cocotb_test_fixture.write({"input_data": "hello world"})
      cocotb_test_fixture.run()
    ```

    The example will assume your toplevel module is `"input_buffer"` and
    it's source file lives in a sibling folder of the `test` folder that
    contains the pytest test function.
    It will create a unique subdirectory under `build_test` that matches
    the path to the module containing the testbench definition and pytest
    test function (those need to be the same).
    This prevents test A overriding the artifacts of test B.
    The name of the subdirectory will be derived from the parameters
    passed via the `parametrize` pytest marker and the top module name.
    In this example this results in folders: `input_buffer_test_input_buffer_x_1`, `input_buffer_test_input_buffer_x_2`, `input_buffer_test_input_buffer_x_3`.

    """

    @wraps(fn)
    async def wrapper(dut):
        build_test_dir = environ["EAI_SIM_TEST_DIR"]
        with open(Path(build_test_dir) / "testdata.json", "r") as f:
            kwargs = json.load(f)
        await fn(dut=dut, **kwargs)

    return wrapper


class CocotbTestFixture:
    """Run cocotb via pytest, inject parameters before and during test execution.

    The fixture will inspect the requesting test function to assume some default values and perform a little bit of setup. Namely this is

    * Use the test function name to determine the dut top module name and the name of its containing source file. These can be overriden inside the test function using `CocotbTestFixture.set_top_module_name()` and `CocotbTestFixture.set_srcs()`. The name will be derived by stripping the `test_` prefix from the test function name.
      The implementation will try to find a vhdl or verilog file under `../{vhdl, verilog}/<name>.{vhd, v}`. Vhdl will take precedence. If no file is found, the initial srcs list will be left empty without raising an exception.
    * It will create a folder to contain test artifacts including waveforms, xml result, testdata json and compiled simulation object files. To avoid collisions, the name of the folder will be derived from the fully qualified test function name (replacing `.` by `/`) and the parameter list provided via pytest parametrization.

    This is not intended to be used directly. Request `cocotb_test_fixture` as a pytest fixture instead.
    """

    def __init__(
        self,
        test_fn: Callable,
        *args: float | int | str,
        **kwargs: float | int | str,
    ) -> None:
        self._test_fn = test_fn
        self._args = args
        self._kwargs = kwargs
        self._build_test_subdir = "<none>"
        self._top_module_name = self._test_fn.__name__.removeprefix("test_")
        self._id = "<none>"
        self._timescale = ("1ps", "1fs")

        def get_parent_module(module: str):
            return ".".join(module.split(".")[:-1])

        self._context_stack = ExitStack()
        self._srcs: list[str] = []

    def setup(self):
        self._create_build_dir()
        self._set_default_srcs()

    def teardown(self):
        self._srcs = []
        self._context_stack.close()

    def _set_default_srcs(self) -> None:
        try:
            src_file = importlib.resources.path(
                ".".join(self._test_fn.__module__.split(".")[:-2] + ["vhdl"]),
                f"{self._top_module_name}.vhd",
            )
            self._srcs = [str(self._context_stack.enter_context(src_file))]
        except (ModuleNotFoundError, FileNotFoundError):
            try:
                src_file = importlib.resources.path(
                    ".".join(self._test_fn.__module__.split(".")[:-2] + ["verilog"]),
                    f"{self._top_module_name}.v",
                )
                self._srcs = [str(self._context_stack.enter_context(src_file))]
            except (ModuleNotFoundError, FileNotFoundError):
                pass

    def _create_build_dir(self):
        build_test_subdir = create_name_for_build_test_subdir(
            self._test_fn, *self._args, **self._kwargs
        )
        build_test_subdir = (
            self._test_fn.__module__.replace(".", "/") + "/" + build_test_subdir
        )

        self._build_test_subdir = build_report_folder_and_testdata(
            build_test_subdir, _get_args(self._test_fn, *self._args, **self._kwargs)
        )

    def write(self, data: dict[str, Any]) -> None:
        with open(Path(self._build_test_subdir) / Path("testdata.json"), "r") as f:
            testdata = json.load(f)
        testdata.update(data)
        with open(Path(self._build_test_subdir) / Path("testdata.json"), "w") as f:
            json.dump(testdata, f)

    def set_top_module_name(self, top_module_name: str) -> None:
        self._top_module_name = top_module_name

    def set_srcs(self, srcs: Iterable[str | Path]):
        self._srcs = list((str(s) for s in srcs))

    def add_srcs(self, *srcs: str | Path) -> None:
        self._srcs.extend((str(s) for s in srcs))

    def set_timescale(self, scale: tuple[str, str]) -> None:
        self._timescale = scale

    def run(self, params, defines):
        environ["EAI_SIM_TEST_DIR"] = str(self._build_test_subdir)

        run_cocotb_sim(
            src_files=self._srcs,
            top_module_name=self._top_module_name,
            cocotb_test_module=self._test_fn.__module__,
            params=params,
            defines=defines,
            build_sim_dir=self._build_test_subdir,
            timescale=self._timescale,
        )


@pytest.fixture
def cocotb_test_fixture(request) -> Iterator[CocotbTestFixture]:
    """Yields the setup CocotbTestFixture and performs necessary clean up after the test run.

    To use the fixture either place add the line
    ```python
    pytest_plugins = "elasticai.creator.testing.cocotb_pytest"
    ```
    to either a conftest.py in the test directory tree or in the
    test module.
    """
    fixture = CocotbTestFixture(request.function, **request.node.callspec.params)
    fixture.setup()
    yield fixture
    fixture.teardown()



from collections.abc import Callable, Iterable, Sequence
from typing import Protocol, TypeVar

_T = TypeVar("_T")


class Shape(Protocol):
    width: int

    def size(self) -> int: ...


class Node(Protocol):
    name: str
    implementation: str
    input_shape: Shape
    output_shape: Shape


def _for_each_produce_multiple_lines(
    fn: Callable[[_T], Iterable[str]],
) -> Callable[[Iterable[_T]], list[str]]:
    def _wrapped(items: Iterable[_T]) -> list[str]:
        result: list[str] = list()
        for item in items:
            result.extend(fn(item))

        return result

    return _wrapped


def _for_each_produce_single_line(
    fn: Callable[[_T], str],
) -> Callable[[Iterable[_T]], list[str]]:
    @_for_each_produce_multiple_lines
    def _wrapped(item: _T) -> Iterable[str]:
        yield fn(item)

    return _wrapped


@_for_each_produce_multiple_lines
def connect_data_signals(
    edge: tuple[str, str, Sequence[tuple[int, int]] | tuple[str, str]],
):
    def get_args(r) -> tuple[int] | tuple[int, int]:
        args = tuple(int(n) for n in r.strip("range(").rstrip(")").split(","))
        if len(args) not in (1, 2):
            raise ValueError(f"Invalid range: {r}")
        return args  # pyright: ignore

    source, dst, wiring = edge

    if len(wiring) >= 1 and isinstance(wiring[0], tuple) and len(wiring[0]) == 2:
        for i, j in wiring:
            yield f"d_in_{dst}({j}) <= d_out_{source}({i});"
        return
    if len(wiring) == 0:
        yield f"d_in_{dst} <= d_out_{source};"
        return
    elif (
        len(wiring) == 2
        and isinstance(wiring[0], str)
        and isinstance(wiring[1], str)
        and wiring[0].startswith("range")
        and wiring[1].startswith("range")
    ):
        src_args = get_args(wiring[0])
        dst_args = get_args(wiring[1])
        if len(src_args) <= 2 and len(dst_args) <= 2:
            if len(src_args) == 1:
                src_args = (0, src_args[0])
            if len(dst_args) == 1:
                dst_args = (0, dst_args[0])
            yield f"d_in_{dst}({dst_args[1] - 1} downto {dst_args[0]}) <= d_out_{source}({src_args[1] - 1} downto {src_args[0]});"
            return
        else:
            wiring = zip(range(*src_args), range(*dst_args))
    for i, j in wiring:  # pyright: ignore
        yield f"d_in_{dst}({j}) <= d_out_{source}({i});"


def _define_signal_vector(name, length):
    return f"signal {name}: std_logic_vector({length}-1 downto 0) := (others => '0');"


@_for_each_produce_single_line
def define_input_data_signals(instance):
    return _define_signal_vector(f"d_in_{instance.name}", instance.input_shape.size())


@_for_each_produce_single_line
def define_output_data_signals(instance: Node):
    return _define_signal_vector(f"d_out_{instance.name}", instance.output_shape.size())


@_for_each_produce_single_line
def instantiate_bufferless(instance: Node):
    return f"{instance.name} : entity work.{instance.implementation}(rtl) port map (x => x_{instance.name}, y => y_{instance.name}, enable => enable);"



from collections.abc import Iterator
from typing import Generic, TypeVar

from elasticai.creator.graph.graph import Graph

T = TypeVar("T")
TP = TypeVar("TP")


class State(Generic[T, TP]):
    def __init__(self, graph: Graph[T]) -> None:
        self.order: dict[T, int] = {n: i for i, n in enumerate(graph.nodes)}
        self.order_back: dict[int, T] = {v: k for k, v in self.order.items()}
        self.core: list[TP | None] = [None for _ in graph.nodes]
        self.in_nodes: list[int] = [0 for _ in graph.nodes]
        self.out_nodes: list[int] = [0 for _ in graph.nodes]
        self.graph: Graph[T] = graph
        self.current_depth = 0

    def remove_pair(self, a: T, b: TP) -> None:
        if self.current_depth == 1:
            self.in_nodes[self.order[a]] = 0
            self.out_nodes[self.order[a]] = 0
        self.core[self.order[a]] = None

    def contains_pair(self, a: T, b: TP) -> bool:
        return self.core[self.order[a]] == b

    def contains_node(self, node: T) -> bool:
        return self.core[self.order[node]] is not None

    def partial_nodes(self) -> set[T]:
        return set(self.iter_nodes())

    def partial_predecessors(self, n: T) -> set[T]:
        return set(self.graph.predecessors[n]).intersection(self.partial_nodes())

    def partial_successors(self, n: T) -> set[T]:
        return set(self.graph.successors[n]).intersection(self.partial_nodes())

    def unseen_nodes(self) -> set[T]:
        depth = self.current_depth
        unseen = set()
        for node, id in self.order.items():
            if not (
                self.in_nodes[id] == depth
                or self.out_nodes[id] == depth
                or self.core[id] is not None
            ):
                unseen.add(node)
        return unseen

    def in_node_successors(self, n: T) -> set[T]:
        nodes: set[T] = set()
        for n in self.graph.successors[n]:
            if self.in_nodes[self.order[n]] == self.current_depth:
                nodes.add(n)
        return nodes

    def in_node_predecessors(self, n: T) -> set[T]:
        nodes: set[T] = set()
        for n in self.graph.predecessors[n]:
            if self.in_nodes[self.order[n]] == self.current_depth:
                nodes.add(n)

        return nodes

    def out_node_successors(self, n: T) -> set[T]:
        nodes: set[T] = set()
        for n in self.graph.successors[n]:
            if self.out_nodes[self.order[n]] == self.current_depth:
                nodes.add(n)
        return nodes

    def out_node_predecessors(self, n: T) -> set[T]:
        nodes: set[T] = set()
        for n in self.graph.predecessors[n]:
            if self.out_nodes[self.order[n]] == self.current_depth:
                nodes.add(n)
        return nodes

    def unseen_successors(self, n: T) -> set[T]:
        return set(self.graph.successors[n]).intersection(self.unseen_nodes())

    def unseen_predecessors(self, n: T) -> set[T]:
        return set(self.graph.predecessors[n]).intersection(self.unseen_nodes())

    def iter_nodes(self) -> Iterator[T]:
        for n in self.order:
            if self.contains_node(n):
                yield n

    def is_complete(self) -> bool:
        return all(n is not None for n in self.core)

    def iter_matched_pairs(self) -> Iterator[tuple[T, TP]]:
        for node in self.graph.nodes:
            matched_node = self.core[self.order[node]]
            if matched_node is not None:
                yield node, matched_node

    def add_pair(self, a: T, b: TP) -> None:
        self.core[self.order[a]] = b
        self.in_nodes[self.order[a]] = self.current_depth
        self.out_nodes[self.order[a]] = self.current_depth

    def restore(self) -> None:
        """reset all changes introduced in the current depth.

        This is responsible for backtracking.
        """
        depth = self.current_depth
        for node_id, match in enumerate(self.core):
            if self.in_nodes[node_id] == depth:
                self.in_nodes[node_id] = 0
            if self.out_nodes[node_id] == depth:
                self.out_nodes[node_id] = 0
        self.current_depth -= 1

    def update_in_nodes(self) -> None:
        """Find out which nodes inside of current match are reachable from outside of current match."""
        depth = self.current_depth
        for n in self.order:
            if not self.contains_node(n):
                for nbr in self.graph.successors[n]:
                    if self.contains_node(nbr):
                        self.in_nodes[self.order[n]] = depth
                        break

    def update_out_nodes(self) -> None:
        """Find out which nodes outside of current match are reachable from the current match."""
        depth = self.current_depth
        for n in self.order:
            if not self.contains_node(n):
                for nbr in self.graph.predecessors[n]:
                    if self.contains_node(nbr):
                        self.out_nodes[self.order[n]] = depth
                        break

    def next_depth(self) -> None:
        self.current_depth += 1
        self.update_inout_nodes()

    def update_inout_nodes(self) -> None:
        self.update_in_nodes()
        self.update_out_nodes()

    def __iter_nodes(self, depth: int, in_out: list[int]) -> Iterator[T]:
        if depth == 0:
            return
        for node, node_id in self.order.items():
            added_at = in_out[node_id]
            if added_at >= depth:
                yield node

    def iter_in_nodes(self) -> Iterator[T]:
        depth = self.current_depth
        yield from self.__iter_nodes(depth, self.in_nodes)

    def iter_out_nodes(self) -> Iterator[T]:
        depth = self.current_depth
        yield from self.__iter_nodes(depth, self.out_nodes)



from elasticai.creator.base_modules.relu import ReLU

from .batchnorm2d import BatchNorm2d
from .conv1d import Conv1d
from .conv2d import Conv2d
from .linear import Linear

__all__ = ["ReLU", "BatchNorm2d", "Conv1d", "Conv2d", "Linear", "parametrized_modules"]
parametrized_modules = [
    "ParametrizedBatchNorm2d",
    "ParametrizedConv1d",
    "ParametrizedConv2d",
    "ParametrizedLinear",
]



from elasticai.creator_plugins.combinatorial.language import Port, VHDLEntity


def test_can_generate_entity():
    expected = [
        "library ieee;",
        "use ieee.std_logic_1164.all;",
        "entity my_entity is",
        "generic (",
        "A : int;",
        "B : positive",
        ");",
        "port (",
        "signal clk : in std_logic;",
        "signal y : out std_logic_vector(3 - 1 downto 0)",
        ");",
        "end entity;",
    ]

    impl = VHDLEntity(
        name="my_entity",
        port=Port(
            inputs=dict(
                clk="std_logic",
            ),
            outputs=dict(y="std_logic_vector(3 - 1 downto 0)"),
        ),
        generics=dict(A="int", B="positive"),
    )
    actual = list(map(str.strip, impl.generate_entity()))
    assert actual == expected



import torch

from elasticai.creator.base_modules.math_operations_torch import TorchMathOperations
from tests.tensor_test_case import TensorTestCase


class TorchMathOperationsTest(TensorTestCase):
    def setUp(self) -> None:
        self.operations = TorchMathOperations()

    def test_quantize(self) -> None:
        a = torch.tensor([0, 1.5, 2.25, 3.123456])
        actual = self.operations.quantize(a)
        self.assertTensorEqual(a, actual)

    def test_add(self) -> None:
        a = torch.tensor([1, 2, 3])
        b = torch.tensor([4, 5, 6])
        actual = self.operations.add(a, b)
        expected = [5, 7, 9]
        self.assertTensorEqual(expected, actual)

    def test_mul(self) -> None:
        a = torch.tensor([1, 2, 3])
        b = torch.tensor([4, 5, 6])
        actual = self.operations.mul(a, b)
        expected = [4, 10, 18]
        self.assertTensorEqual(expected, actual)

    def test_matmul(self) -> None:
        a = torch.tensor([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        b = torch.tensor([[1], [2], [3]])
        actual = self.operations.matmul(a, b)
        expected = [[14], [32], [50]]
        self.assertTensorEqual(expected, actual)



from typing import Any

from pytest import fixture

from elasticai.creator.graph.base_graph import BaseGraph
from elasticai.creator.ir2vhdl import (
    Edge,
    Implementation,
    Instance,
    LogicSignal,
    LogicVectorSignal,
    NullDefinedLogicSignal,
    PortMap,
    Shape,
    Signal,
    VhdlNode,
    vhdl_node,
)


@fixture
def impl() -> Implementation:
    impl: Implementation[VhdlNode, Edge] = Implementation(
        name="conv1",
        type="conv",
        data={"a": 1},
    )
    impl.add_node(
        VhdlNode(name="x", data=dict(type="y")),
    )
    return impl


@fixture
def data() -> dict[str, Any]:
    return {
        "name": "conv1",
        "type": "conv",
        "nodes": {
            "x": {"type": "y"},
        },
        "edges": {},
        "a": 1,
    }


def test_store_as_dict(data, impl):
    assert data == impl.as_dict()


def test_load_from_dict(data):
    assert data == Implementation(graph=BaseGraph()).load_from_dict(data).as_dict()


def test_can_access_attributes_of_vhdl_node():
    n = VhdlNode(name="a", data={"type": "b", "implementation": "c", "stride": 2})
    n.input_shape = Shape(1)
    n.output_shape = Shape(2)
    assert n.attributes["stride"] == 2


class TestInstance:
    def test_instantiate(self):
        n = vhdl_node(
            name="my_component",
            type="my_type",
            implementation="my_implementation",
            input_shape=Shape(2, 8),
            output_shape=Shape(2),
            attributes={},
        )
        my_instance = Instance(
            n,
            dict(data_depth="8"),
            dict(
                clk=NullDefinedLogicSignal("clk"),
                valid_in=LogicSignal("valid_in"),
                d_in=LogicVectorSignal("data_in", 16),
            ),
        )
        expected = (
            "my_component: entity work.my_implementation(rtl) ",
            "generic map (",
            "  DATA_DEPTH => 8",
            "  )",
            "  port map (",
            "    clk => clk,",
            "    valid_in => valid_in_my_component,",
            "    d_in => data_in_my_component",
            "  );",
        )

        assert expected == tuple(my_instance.instantiate())

    def test_can_instantiate_with_two_generics(self):
        n = vhdl_node(
            name="my_component",
            type="my_type",
            implementation="my_implementation",
            input_shape=Shape(2, 8),
            output_shape=Shape(2),
            attributes={},
        )
        my_instance = Instance(
            n,
            dict(data_depth="8", other="10"),
            dict(),
        )
        expected = (
            "my_component: entity work.my_implementation(rtl) ",
            "generic map (",
            "  DATA_DEPTH => 8,",
            "  OTHER => 10",
            "  )",
            "  port map (",
            "  );",
        )

        assert expected == tuple(my_instance.instantiate())


class TestLogicSignal:
    def test_can_create_from_code_with_default_val(self):
        code = "signal clk: std_logic := '0';"
        signal = Signal.from_code(code)
        assert signal.name == "clk"

    def test_can_create_from_code(self):
        code = "signal clk : std_logic;"
        signal = Signal.from_code(code)
        assert signal.name == "clk"

    def test_can_define_signal(self):
        code = ("signal clk : std_logic := '0';",)
        signal = LogicSignal("clk")
        assert code == tuple(signal.define())

    def test_make_instance_specific(self):
        code = ("signal clk_my_component : std_logic := '0';",)
        signal = LogicSignal("clk")
        signal = signal.make_instance_specific("my_component")
        assert code == tuple(signal.define())


class TestLogicVectorSignal:
    def test_can_create_from_code_with_default_val(self):
        code = 'signal d_in: std_logic_vector(3 downto 0) := "0000";'
        signal = Signal.from_code(code)
        assert signal.name == "d_in"
        assert signal.width == 4

    def test_can_create_from_signal_with_arith_expr(self):
        code = 'signal d_in: std_logic_vector(3 - 1 downto 0) := "000";'
        signal = Signal.from_code(code)
        assert signal.width == 3

    def test_can_define(self):
        code = ("signal d_in : std_logic_vector(4 - 1 downto 0) := (others => '0');",)
        signal = LogicVectorSignal("d_in", 4)
        assert code == tuple(signal.define())


class TestPortMap:
    def test_can_get_as_dict(self):
        p = PortMap(
            dict(
                d_in=LogicVectorSignal("d_in", 4),
                clk=LogicSignal("clk"),
            )
        )
        assert p.as_dict() == {
            "d_in": "signal d_in : std_logic_vector(4 - 1 downto 0) := (others => '0');",
            "clk": "signal clk : std_logic := '0';",
        }

    def test_can_get_from_dict(self):
        p = PortMap(
            dict(
                d_in=LogicVectorSignal("d_in", 4),
                clk=LogicSignal("clk"),
            )
        )
        assert p == PortMap.from_dict(
            {
                "d_in": "signal d_in : std_logic_vector(4 - 1 downto 0) := (others => '0');",
                "clk": "signal clk : std_logic := '0';",
            }
        )



from __future__ import annotations

import copy
import warnings
from collections.abc import Iterable, Iterator, Mapping, Set
from typing import Hashable, Self, TypeVar, cast

from .graph import Graph

T = TypeVar("T", bound=Hashable)  # noqa: F821


# we have to keep this bound to T, until we move to python 3.13 with default type arguments
class BaseGraph(Graph[T]):
    """All iterators in this class are guaranteed to have a fixed
    order. That means, the order of iteration is only allowed
    to change when the data Tucture is altered. If the
    content of two GraphDelegates is the same and they were built
    in the same way, then their iteration order is the same as well.

    :::{caution}
    This class is not thread-safe.
    :::

    :::{note}
    We are not providing methods for removal of nodes or edges on purpose.
        If you need to remove nodes or edges, you should create a new GraphDelegate.
        Manipulation of the graph should usually be done in a dedicated build phase.
    :::
    """

    def __init__(self) -> None:
        """We keep successor and predecessor nodes just to allow for easier implementation.
        Currently, this implementation is not optimized for performance.
        """
        self._successors: dict[T, set[T]] = dict()
        self._predecessors: dict[T, set[T]] = dict()

    @property
    def successors(self) -> Mapping[T, set[T]]:
        return self._successors.keys().mapping

    @property
    def predecessors(self) -> Mapping[T, set[T]]:
        return self._predecessors.keys().mapping

    @staticmethod
    def from_dict(d: dict[T, Iterable[T]]):
        g: Graph[T] = BaseGraph()
        for node, successors in d.items():
            g.add_node(node)
            for s in successors:
                g.add_edge(node, s)
        return g

    def as_dict(self) -> dict[T, set[T]]:
        return self._successors.copy()

    def add_edge(self, _from: T, _to: T) -> Self:
        self.add_node(_from)
        self.add_node(_to)
        self.predecessors[_to].add(_from)
        self.successors[_from].add(_to)
        return self

    def add_node(self, node: T) -> Self:
        if node not in self.predecessors:
            self._predecessors[node] = set()
        if node not in self.successors:
            self._successors[node] = set()
        return self

    @property
    def nodes(self) -> Set[T]:
        return self._successors.keys()

    def iter_nodes(self) -> Iterator[T]:
        """Iterator over nodes in a fixed but unspecified order."""
        yield from self.predecessors.keys()

    def iter_edges(self) -> Iterator[tuple[T, T]]:
        """Iterator over edges in a fixed but unspecified order."""
        for _from, _tos in self.successors.items():
            for _to in _tos:
                yield _from, _to

    def get_edges(self) -> Iterator[tuple[T, T]]:
        """Iterator over edges in a fixed but unspecified order."""
        warnings.warn(
            "get_edges() is deprecated, use iter_edges() instead",
            DeprecationWarning,
            stacklevel=2,
        )
        yield from self.iter_edges()

    def get_successors(self, node: T) -> Iterator[T]:
        """Iterator over node successors in a fixed but unspecified order."""
        yield from self.successors[node]

    def get_predecessors(self, node: T) -> Iterator[T]:
        """Iterator over node predecessors in a fixed but unspecified order."""
        yield from self.predecessors[node]

    def new(self: Self) -> Self:
        return cast(Self, BaseGraph())

    def copy(self: Self) -> Self:
        g = self.new()
        g._predecessors = copy.deepcopy(self._predecessors)
        g._successors = copy.deepcopy(self._successors)
        return g



from collections.abc import Iterator

from elasticai.creator.ir2vhdl import Implementation

from .vhdl_nodes.node_factory import InstanceFactoryForCombinatorial
from .wiring import (
    connect_data_signals,
)


def wrap_node(n):
    return InstanceFactoryForCombinatorial(n)


def build_declarations_for_combinatorial(impl: Implementation) -> Iterator[str]:
    nodes = tuple(wrap_node(n) for n in impl.nodes.values())
    signal_defs = (line for n in nodes for line in n.define_signals())
    yield from signal_defs


def build_instantiations_for_combinatorial(impl: Implementation) -> Iterator[str]:
    nodes = tuple(wrap_node(n) for n in impl.nodes.values())
    instances = tuple(n for n in nodes if n.name not in ("input", "output"))
    instantiations = (line for n in instances for line in n.instantiate())
    yield from instantiations


def build_data_signal_connections_for_combinatorial(
    impl: Implementation,
) -> Iterator[str]:
    connections = tuple((e.src, e.dst, e.src_dst_indices) for e in impl.edges.values())
    data_signal_connections = connect_data_signals(connections)
    yield from data_signal_connections


def wrap_in_architecture(name, declarations, definitions):
    def _indent(line):
        return f"  {line}"

    yield f"architecture rtl of {name} is"
    yield from map(_indent, declarations)
    yield "begin"
    yield from map(
        _indent,
        (
            "d_in_input <= d_in;",
            "d_out_input <= d_in_input;",
            "d_out_output <= d_in_output;",
            "d_out <= d_out_output;",
        ),
    )
    yield from map(_indent, definitions)
    yield "end architecture;"



from collections.abc import Callable

from .attribute import Attribute
from .attributes_descriptor import AttributesDescriptor
from .ir_data_meta import IrDataMeta


class IrData(metaclass=IrDataMeta, create_init=False):
    """Convenience class for creating new Ir data classes.

    To create a new Ir data class, inherit from `IrData` and define your required fields.
    Supported field types are

    - ['RequiredField'](#elasticai.creator.ir.required_field.RequiredField)
    - [`ReadOnlyField`](#elasticai.creator.ir.required_field.ReadOnlyField)
    - [`SimpleField`](#elasticai.creator.ir.required_field.SimpleField)
    - [`StaticMethodField`](#elasticai.creator.ir.required_field.static_required_field)
    - [`ReadOnlyMethodField`](#elasticai.creator.ir.required_field.read_only_field)

    Examples:

    We can define a new class `MyNewIrData` with the two required fields `name` and `length` like this:

    ```python
    from elasticai.creator.ir import IrData, SimpleRequiredField, RequiredField

    class MyNewIrData(IrData):
        name: SimpleRequiredField[str] = SimpleRequiredField()  # <1>
        length: RequiredField[str, int] = RequiredField(str, int)  # <2>

    d = MyNewIrData({'name': 'my name', 'length': '12'})
    d.name = 'new name'  # <3>
    d.length += 3  # <4>
    ```

    1. We define a simple required field. Data will be stored as string and can be accessed as string.
    2. Opposed to 1. this field will store data as a string but every read or write access will convert
    length to an integer. The constructor `str` will be called when writing a value to the field and `int`
    will be called on read.
    3. `'new name'` will be stored as is inside `d`'s data dictionary.
    4  `'12'` will be converted to `12` before adding `3` to it, the subsequent write operation will convert `15`
    to the string `'15'` again.

    IMPORTANT: Currently we assume that calling the conversion functions for a field will not result in any side effects.
    IMPORTANT: While it is ok to extend the available field types by inheritance the *type annotations*
      still need to refer to one of the provided field types, because they need to be analyzed by the metaclass.
    """

    _fields: dict[str, type]  # only here for type checkers

    attributes: AttributesDescriptor = AttributesDescriptor()

    def __init__(self, data: dict[str, Attribute]) -> None:
        self.data = data

    def get_missing_required_fields(self) -> dict[str, type]:
        missing: dict[str, type] = {}
        for name, _type in self._fields.items():
            if name not in self.data:
                missing[name] = _type
        return missing

    def __repr__(self) -> str:
        return f"{self.__module__}.{self.__class__.__qualname__}({self.data})"

    def __eq__(self, o: object) -> bool:
        if self is o:
            return True
        if hasattr(o, "data"):
            return o.data == self.data

        return False


def ir_data_class(cls) -> Callable[[dict[str, Attribute]], IrData]:
    return type(cls.__name__, (cls, IrData), {})



import re

import pytest

from elasticai.creator.vhdl.auto_wire_protocols.autowiring import (
    AutoWirer,
    AutoWiringProtocolViolation,
    DataFlowNode,
)


def _wire(
    top: DataFlowNode, graph: tuple[DataFlowNode, ...]
) -> dict[tuple[str, str], tuple[str, str]]:
    _autowirer = AutoWirer()
    _autowirer.wire(top, graph)
    return _autowirer.connections()


def _build_expected_connections(spec: str):
    regex = r"\s*(\w+)\s*,\s*(\w+)\s*:\s*(\w+)\s*,\s*(\w+)\s*\n*"
    result = {}
    for match in re.finditer(regex, spec):
        dst_node, dst_signal, source_node, source_signal = match.groups()
        result[(dst_node, dst_signal)] = (source_node, source_signal)
    return result


def test_wire_buffered_entity_to_top_module() -> None:
    child = DataFlowNode.buffered("child")
    connections = _wire(top=DataFlowNode.top("top"), graph=(child,))
    expected_connections = {
        ("child", "x"): ("top", "x"),
        ("child", "y_address"): ("top", "y_address"),
        ("child", "clock"): ("top", "clock"),
        ("child", "enable"): ("top", "enable"),
        ("top", "y"): ("child", "y"),
        ("top", "x_address"): ("child", "x_address"),
        ("top", "done"): ("child", "done"),
    }
    assert connections == expected_connections


def test_wire_two_buffered_entities() -> None:
    a = DataFlowNode.buffered("a")
    b = DataFlowNode.buffered("b")
    top = DataFlowNode.top("top")
    expected_connections = {
        ("a", "clock"): ("top", "clock"),
        ("a", "x"): ("top", "x"),
        ("a", "enable"): ("top", "enable"),
        ("top", "x_address"): ("a", "x_address"),
        ("a", "y_address"): ("b", "x_address"),
        ("b", "clock"): ("top", "clock"),
        ("b", "x"): ("a", "y"),
        ("b", "y_address"): ("top", "y_address"),
        ("b", "enable"): ("a", "done"),
        ("top", "y"): ("b", "y"),
        ("top", "done"): ("b", "done"),
    }
    connections = _wire(top=top, graph=(a, b))
    assert connections == expected_connections


def test_wire_unbuffered_entity() -> None:
    a = DataFlowNode.unbuffered("a")
    expected_connections = {
        ("top", "y"): ("a", "y"),
        ("top", "done"): ("top", "enable"),
        ("top", "x_address"): ("top", "y_address"),
        ("a", "enable"): ("top", "enable"),
        ("a", "x"): ("top", "x"),
        ("a", "clock"): ("top", "clock"),
    }
    connections = _wire(top=DataFlowNode.top("top"), graph=(a,))
    assert connections == expected_connections


def test_wire_buffered_and_unbuffered() -> None:
    a = DataFlowNode.buffered("a")
    b = DataFlowNode.unbuffered("b")
    top = DataFlowNode.top("top")
    spec = """
        top, x_address : a, x_address
        a, clock : top, clock
        a, enable : top, enable
        a, y_address : top, y_address
        a, x : top, x
        b, enable : a, done
        b, clock : top, clock
        b, x : a, y
        top, y: b, y
        top, done: a, done
        """
    expected_connections = _build_expected_connections(spec)
    connections = _wire(top=top, graph=(a, b))
    assert connections == expected_connections


def test_wiring_unknown_signals_yields_error() -> None:
    a = DataFlowNode(name="a", dsts=("signal_a",), sources=tuple())
    with pytest.raises(AutoWiringProtocolViolation):
        _wire(top=DataFlowNode.top("top"), graph=(a,))



from collections.abc import Callable, Iterable
from typing import ParamSpec

import elasticai.creator.plugin as _pl
from elasticai.creator.function_utils import FunctionDecorator
from elasticai.creator.graph import dfs_iter
from elasticai.creator.ir import RequiredField
from elasticai.creator.ir2vhdl import (
    Edge,
    Implementation,
    LoweringPass,
    Shape,
    edge,
    vhdl_node,
)
from elasticai.creator.ir2vhdl import (
    VhdlNode as Node,
)
from elasticai.creator_plugins.grouped_filter import FilterParameters

P = ParamSpec("P")


def _type_handler_fn(
    name: str, fn: Callable[[Implementation], Implementation]
) -> _pl.PluginSymbolFn[LoweringPass, [Implementation], Implementation]:
    def load_into(lower: LoweringPass[Implementation, Implementation]):
        lower.register(name, fn)

    return _pl.make_plugin_symbol(load_into, fn)


def _iterable_type_handler_fn(
    name: str, fn: Callable[[Implementation], Iterable[Implementation]]
) -> _pl.PluginSymbolFn[LoweringPass, [Implementation], Iterable[Implementation]]:
    def load_into(lower: LoweringPass[Implementation, Implementation]):
        lower.register_iterable(name, fn)

    return _pl.make_plugin_symbol(load_into, fn)


_type_handler = FunctionDecorator(_type_handler_fn)
_iterable_type_handler = FunctionDecorator(_iterable_type_handler_fn)


class _FilterNode(Node):
    filter_parameters: RequiredField[dict, FilterParameters] = RequiredField(
        set_convert=lambda x: x.as_dict(), get_convert=FilterParameters.from_dict
    )


def append_counter_suffix_before_construction(
    fn: Callable[P, Node],
) -> Callable[P, Node]:
    counters: dict[str, int] = {}

    def construct(*args: P.args, **kwargs: P.kwargs) -> Node:
        nonlocal counters
        if "name" in kwargs:
            name = kwargs["name"]
            if not isinstance(name, str):
                raise TypeError("expected `name` to be of type str")
        elif isinstance(args[0], str):
            name = args[0]
        else:
            raise TypeError("missing positional arg `name`")

        count = counters.get(name, 0)
        counters[name] = count + 1
        kwargs["name"] = f"{name}_i{count}"
        node = fn(*args, **kwargs)
        return node

    return construct


class _Sequential:
    def __init__(self, name: str):
        self._impl: Implementation[Node, Edge] = Implementation(
            name=name,
            type="clocked_combinatorial",
        )
        self._last_node: Node | None = None
        self._last_filter_parameters: None | FilterParameters = None
        self._counting_node_constructor = append_counter_suffix_before_construction(
            vhdl_node
        )
        self._num_registers: int = 0
        self._last_stride = 1

    def add_input(self, shape: Shape):
        self._impl.add_node(
            vhdl_node(
                name="input",
                type="input",
                implementation="",
                output_shape=shape,
                input_shape=shape,
            )
        )
        self._last_node = self._impl.nodes["input"]
        self._update_last_filter_params()

    def _need_shift_register(self, params: FilterParameters) -> bool:
        if self._last_node is None:
            return False
        consuming_more_than_last_node_produces = (
            self._last_node.output_shape.width < params.kernel_size
        )
        return consuming_more_than_last_node_produces

    def _need_sliding_window(self, params: FilterParameters) -> bool:
        if self._last_node is None:
            return False
        consuming_less_than_last_node_produces = (
            self._last_node.output_shape.width > params.kernel_size
        )
        return consuming_less_than_last_node_produces

    def filter(self, n: Node):
        node = _FilterNode(n.name, n.data)
        attributes = n.attributes
        params = node.filter_parameters
        if "top_stride" not in self._impl.attributes:
            self._impl.data.update(
                {
                    "top_stride": params.in_channels * params.stride,
                }
            )
        if self._need_shift_register(params):
            old_params = self._last_filter_parameters
            self.strided_shift_register(
                output_shape=(
                    params.in_channels,
                    params.kernel_size,
                ),
                stride=old_params.stride,
            )
        elif self._need_sliding_window(params):
            self._sliding_window(params)
        elif self._last_node is not None:
            pass
        else:
            raise ValueError("expected last node to be not None")
        self._append_node(
            name=n.name,
            type="unclocked_combinatorial",
            implementation=n.implementation,
            output_shape=Shape(attributes["filter_parameters"]["out_channels"], 1),
            attributes=attributes,
            node_fn=vhdl_node,
        )

    def _sliding_window(self, params: FilterParameters) -> None:
        self._append_static(
            "sliding_window",
            "sliding_window",
            output_shape=Shape(
                params.in_channels,
                params.kernel_size,
            ),
        )

    def _update_last_filter_params(self):
        new_node = self._last_node
        if "filter_parameters" in new_node.data:
            params = _FilterNode(new_node.name, new_node.data).filter_parameters

            self._last_filter_parameters is not None
            if params.kernel_size == 1:
                self._last_filter_parameters = FilterParameters(
                    kernel_size=params.kernel_size,
                    in_channels=params.in_channels,
                    out_channels=params.out_channels,
                    groups=params.groups,
                    stride=self._last_filter_parameters.stride * params.stride,
                    input_size=params.input_size,
                    output_size=params.output_size,
                )
            else:
                self._last_filter_parameters = params
        elif self._last_filter_parameters is None:
            self._last_filter_parameters = FilterParameters(
                kernel_size=1,
                in_channels=new_node.input_shape.depth,
                out_channels=new_node.output_shape.depth,
            )

    def _append_node(
        self,
        name: str,
        output_shape: Shape,
        type: str,
        implementation: str,
        node_fn,
        attributes=None,
    ):
        old_node = self._last_node
        if old_node is None:
            raise Exception("no input node")
        input_shape = old_node.output_shape
        new_node = node_fn(
            name=name,
            input_shape=input_shape,
            output_shape=output_shape,
            type=type,
            implementation=implementation,
            attributes=attributes if attributes is not None else {},
        )
        self._impl.add_node(new_node)
        self._last_node = new_node
        self._update_last_filter_params()
        if old_node is not None:
            self._impl.add_edge(
                edge(
                    src=old_node.name,
                    dst=new_node.name,
                    src_dst_indices=tuple(),
                )
            )

    def _append_static(
        self, name: str, implementation: str, output_shape: Shape, **kwargs
    ):
        self._append_node(
            name=name,
            output_shape=output_shape,
            implementation=implementation,
            type=implementation,
            attributes=kwargs,
            node_fn=self._counting_node_constructor,
        )

    def shift_register(self, name: str, output_shape: Shape):
        self._append_static(
            name=name,
            output_shape=output_shape,
            implementation="shift_register",
        )

    def strided_shift_register(self, output_shape: tuple[int, int], stride: int):
        self._num_registers += 1
        if stride > 1:
            self._append_static(
                name="striding_shift_register",
                output_shape=Shape(*output_shape),
                implementation="striding_shift_register",
                generic_map=dict(stride=stride),
            )
        else:
            self._append_static(
                name="shift_register",
                output_shape=Shape(*output_shape),
                implementation="shift_register",
                generic_map=dict(),
            )

    def input(self, impl: Implementation, input_node: str) -> None:
        input_shape = self._determine_required_input_shape(impl, input_node)
        self._impl.data["top_kernel_size"] = input_shape.size()
        self.add_input(input_shape)

    def _determine_required_input_shape(
        self, impl: Implementation, input_node: str
    ) -> Shape:
        first_node_after_input = tuple(impl.successors(input_node).values())[0]
        match first_node_after_input.type:
            case "filter":
                n = _FilterNode(
                    first_node_after_input.name, first_node_after_input.data
                )
                return Shape(
                    n.filter_parameters.in_channels, n.filter_parameters.kernel_size
                )
            case _:
                return impl.nodes[input_node].output_shape

    def set_runtime_input_shape(self, s: Shape) -> None:
        self._impl.data["runtime_input_shape"] = s.to_tuple()

    def set_runtime_output_shape(self, s: Shape) -> None:
        self._impl.data["runtime_output_shape"] = s.to_tuple()

    def get_impl(self) -> Implementation:
        return self._impl

    def output(self, n):
        if self._last_node.output_shape.width < n.input_shape.width:
            self.shift_register("shift_register", n.output_shape)
        self._append_node(
            name="output",
            output_shape=self._last_node.output_shape,
            type="output",
            implementation="",
            node_fn=vhdl_node,
        )


@_type_handler
def sequential(impl: Implementation) -> Implementation:
    seq = _Sequential(impl.name)

    def iter_nodes():
        for input_node in impl.nodes.values():
            if input_node.type == "input":
                break

        def iterator():
            yield from dfs_iter(impl.successors, input_node.name)

        return impl.get_node_mapping(iterator)

    for n in iter_nodes().values():
        match n.type:
            case "filter":
                seq.filter(n)
            case "input":
                seq.set_runtime_input_shape(n.input_shape)
                seq.input(impl, n.name)
            case "output":
                seq.set_runtime_output_shape(n.output_shape)
                seq.output(n)
            case _:
                raise Exception(
                    f"Can't handle unknown type {n.type} during generation of time multiplexed sequential"
                )

    return seq.get_impl()


@_iterable_type_handler
def network(impl: Implementation) -> Iterable[Implementation]:
    network = sequential(impl)
    network.attributes["top_kernel_size"]
    network.attributes["top_stride"]
    input_shape = network.nodes["input"].input_shape
    output_shape = network.nodes["output"].output_shape
    input_width, input_depth = input_shape
    output_width, output_depth = output_shape

    skeleton: Implementation = Implementation(name="skeleton", type="skeleton")
    skeleton.data["generic_map"] = {
        "DATA_IN_WIDTH": str(input_width),
        "DATA_IN_DEPTH": str(input_depth),
        "DATA_OUT_WIDTH": str(output_width),
        "DATA_OUT_DEPTH": str(output_depth),
    }

    buffered_network_wrapper: Implementation = Implementation(
        name="buffered_network_wrapper",
        type="buffered_network_wrapper",
    )
    return network, skeleton, buffered_network_wrapper



import cocotb
import pytest

from elasticai.creator.testing import (
    eai_testbench,
)

pytest_plugins = "elasticai.creator.testing.cocotb_pytest"


@cocotb.test()
@eai_testbench
async def my_testbench(dut, x, input_data):
    assert x == dut.X.value
    assert input_data == [1, 2]


@pytest.mark.simulation
@pytest.mark.parametrize(["x"], [(i,) for i in (1, 2, 3)])
def test_my_testbench(cocotb_test_fixture, x):
    additional_input_data = [1, 2]
    cocotb_test_fixture.write({"input_data": additional_input_data})
    cocotb_test_fixture.run(params={"X": x}, defines={})



from elasticai.creator.base_modules.prelu import PReLU as PReLUBase

from .precomputed_module import PrecomputedModule


class PReLU(PrecomputedModule):
    def __init__(
        self,
        total_bits: int,
        frac_bits: int,
        num_steps: int,
        sampling_intervall: tuple[float, float] = (-float("inf"), float("inf")),
    ) -> None:
        """Quantized Activation Function for Adaptive SiLU
        :param total_bits:          Total number of bits
        :param frac_bits:           Fraction of bits
        :param num_steps:           Number of LUT size / total steps in LUT inbetween the sampling interval
        :param sampling_intervall:  Floating tuple with input sampling interval (Note default is [-inf, inf] will be transformed into [-2.0, 1.75] for FxP(4, 2))
        """
        super().__init__(
            base_module=PReLUBase(),
            total_bits=total_bits,
            frac_bits=frac_bits,
            num_steps=num_steps,
            sampling_intervall=sampling_intervall,
        )



import torch

from elasticai.creator.arithmetic import (
    FxpArithmetic,
    FxpParams,
)
from elasticai.creator.nn.fixed_point.math_operations import MathOperations
from tests.tensor_test_case import TensorTestCase


class FixedPointMathOperationsTest(TensorTestCase):
    def setUp(self) -> None:
        self.operations = MathOperations(
            config=FxpArithmetic(FxpParams(total_bits=4, frac_bits=2, signed=True))
        )

    def test_add(self) -> None:
        a = torch.tensor([-0.25, 0.5, 1.0])
        b = torch.tensor([-1.5, 1.0, 1.5])
        actual = self.operations.add(a, b)
        expected = [-1.75, 1.5, 1.75]
        self.assertTensorEqual(expected, actual)

    def test_matmul(self) -> None:
        a = torch.tensor([[-2.0, -1.75, -1.5], [-0.25, 0.0, 0.25], [1.25, 1.5, 1.75]])
        b = torch.tensor([[-0.25], [0.5], [0.25]])
        actual = self.operations.matmul(a, b)
        expected = [[-0.75], [0.0], [0.75]]
        self.assertTensorEqual(expected, actual)

    def test_mul(self) -> None:
        a = torch.tensor([-0.5, 1.5, 0.5])
        b = torch.tensor([0.5, 1.5, 1.2])
        actual = self.operations.mul(a, b)
        expected = [-0.25, 1.75, 0.5]
        self.assertTensorEqual(expected, actual)

    def test_quantize_clamps_minus5_to_minus2(self) -> None:
        a = torch.tensor([-5.0])
        actual = self.operations.quantize(a)
        expected = [-2.0]
        self.assertTensorEqual(expected, actual)

    def test_quantize_rounds_1_1_to_1_0(self) -> None:
        a = torch.tensor([1.1])
        actual = self.operations.quantize(a)
        expected = [1.0]
        self.assertTensorEqual(expected, actual)

    def test_round(self) -> None:
        a = torch.tensor([-0.5, 1.5, 0.5, 0.65])
        actual = self.operations.round(a)
        expected = [-0.5, 1.5, 0.5, 0.75]
        self.assertTensorEqual(expected, actual)

    def test_quantize(self) -> None:
        a = torch.tensor([-0.5, 1.5, 0.5, 0.65])
        actual = self.operations.quantize(a)
        expected = [-0.5, 1.5, 0.5, 0.5]
        self.assertTensorEqual(expected, actual)



import warnings

from elasticai.creator.ir2vhdl import Instance, VhdlNode

from .clocked_combinatorial import ClockedInstance
from .node_factory import (
    InstanceFactoryForCombinatorial,
)


@InstanceFactoryForCombinatorial.register
def shift_register(node: VhdlNode) -> Instance:
    return ShiftRegister(node)


def _check_input_output_shape_compatibility(node):
    input_shape, output_shape = node.input_shape, node.output_shape
    if output_shape.size() % input_shape.size() != 0:
        raise ValueError(
            "Found incompatible input output shapes for shift_register. Total output size has to be an integer multiple of total input size, but found output={output} and input={input}.".format(
                output=output_shape, input=input_shape
            )
        )
    if output_shape.depth != input_shape.depth:
        warnings.warn(
            'Detected mismatching input output shapes for shift_register for node "{}". Width of output and input shape should usually be equal, but found output={} and input={}.'.format(
                node.name, output_shape, input_shape
            ),
            stacklevel=3,
        )


class ShiftRegister(ClockedInstance):
    _logic_signals_with_default_suffix = ("valid_in", "valid_out")

    def __init__(self, node: VhdlNode):
        _check_input_output_shape_compatibility(node)
        data_width = node.input_shape.size()
        output_width = node.output_shape.size()
        if output_width % data_width != 0:
            raise ValueError(
                "incompatible shapes input: {}  output: {}".format(
                    node.input_shape, node.output_shape
                )
            )
        num_points = output_width // data_width
        super().__init__(
            node,
            input_width=data_width,
            output_width=output_width,
            generic_map={"data_width": data_width, "num_points": num_points},
        )



import pytest

from elasticai.creator.file_generation.savable import Path
from elasticai.creator.nn.sequential.design import Sequential
from elasticai.creator.vhdl.auto_wire_protocols.port_definitions import create_port
from elasticai.creator.vhdl.design.design import Design
from elasticai.creator.vhdl.design.ports import Port


class DummyDesign(Design):
    def __init__(
        self, name: str, x_width: int, y_width: int, x_count: int = 0, y_count: int = 0
    ) -> None:
        super().__init__(name)
        self._port = create_port(
            x_width=x_width, y_width=y_width, x_count=x_count, y_count=y_count
        )

    @property
    def port(self) -> Port:
        return self._port

    def save_to(self, destination: Path) -> None:
        pass


class TestSequentialSignalWidthsAreDerivedFromSubdesigns:
    @pytest.fixture
    def port_of_mixed_sequential(self) -> Port:
        return Sequential(
            [
                DummyDesign("dd_0", x_width=4, y_width=3),
                DummyDesign("dd_1", x_width=3, x_count=2, y_width=6, y_count=3),
                DummyDesign("dd_2", x_width=6, y_width=2),
            ],
            name="sequential_0",
        ).port

    def test_x_width_matches(self, port_of_mixed_sequential: Port) -> None:
        assert port_of_mixed_sequential["x"].width == 4

    def test_y_width_matches(self, port_of_mixed_sequential: Port) -> None:
        assert port_of_mixed_sequential["y"].width == 2



__all__ = [
    "RequiredField",
    "SimpleRequiredField",
    "IrData",
    "Edge",
    "Node",
    "edge",
    "node",
    "LoweringPass",
    "Lowerable",
    "Implementation",
    "Attribute",
    "static_required_field",
    "read_only_field",
    "StaticMethodField",
    "ReadOnlyField",
    "GraphProtocol",
    "Rewriter",
    "PatternRuleSpec",
    "RemappedSubImplementation",
    "ReadOnlyDataGraph",
    "Rule",
    "CompositeRule",
    "PatternRule",
    "Pattern",
    "StdPattern",
]
from .base import (
    Attribute,
    IrData,
    ReadOnlyField,
    RequiredField,
    SimpleRequiredField,
    StaticMethodField,
    read_only_field,
    static_required_field,
)
from .core import (
    Edge,
    Implementation,
    Lowerable,
    LoweringPass,
    Node,
    edge,
    node,
)
from .core import (
    Graph as GraphProtocol,
)
from .rewriting import (
    CompositeRule,
    Pattern,
    PatternRuleSpec,
    ReadOnlyDataGraph,
    Rule,
    StdPattern,
)



from pathlib import Path

from elasticai.creator.ir2vhdl.testing import run_vunit_vhdl_testbenches

if __name__ == "__main__":
    run_vunit_vhdl_testbenches(
        ["elasticai.creator_plugins.counter"], test_dir=Path(__file__).parent
    )



from typing import Any

from torch import Tensor
from torch.nn import BatchNorm2d as TorchBatchNorm2d
from torch.nn import Module
from torch.nn.utils import parametrize as P


class BatchNorm2d(TorchBatchNorm2d):
    """A BatchNorm2d.
    The output of the batchnorm is fake quantized. The weights and bias are fake quantized during initialization.
    Make sure that math_ops is a module where all needed tensors are part of it,
    so they can be moved to the same device.
    Make sure that weight_quantization and bias_quantization are modules that implement the forward function.
    If you want to quantize during initialization or only apply quantized updates make sure to use a quantized optimizer
    and implement the right_inverse method for your module.
    """

    def __init__(
        self,
        math_ops: Module,
        weight_quantization: Module,
        bias_quantization: Module,
        num_features: int,
        eps: float = 1e-5,
        momentum: float = 0.1,
        affine: bool = True,
        track_running_stats: bool = True,
        device: Any = None,
        dtype: Any = None,
    ) -> None:
        super().__init__(
            num_features=num_features,
            eps=eps,
            momentum=momentum,
            affine=affine,
            track_running_stats=track_running_stats,
            device=device,
            dtype=dtype,
        )
        P.register_parametrization(self, "weight", weight_quantization)
        P.register_parametrization(self, "bias", bias_quantization)
        self.add_module("math_ops", math_ops)

    def forward(self, x: Tensor) -> Tensor:
        return self.math_ops(super().forward(x))



from os.path import exists
from pathlib import Path

import pytest
import torch

from elasticai.creator.arithmetic import FxpArithmetic, FxpParams
from elasticai.creator.file_generation import find_project_root
from elasticai.creator.file_generation.on_disk_path import OnDiskPath
from elasticai.creator.testing import (
    build_report_folder_and_testdata,
    run_cocotb_sim,
)
from elasticai.creator.vhdl.shared_designs.rom.design import Rom


@pytest.mark.simulation
@pytest.mark.parametrize(
    "total_bits, frac_bits, features_in", [(4, 2, 8), (8, 2, 12), (10, 4, 63)]
)
def test_build_test_rom(
    total_bits: int,
    frac_bits: int,
    features_in: int,
) -> None:
    file_name = f"TestROM_{total_bits}_{frac_bits}_{features_in}"

    # Build pattern and write into file
    fxp = FxpArithmetic(FxpParams(total_bits=total_bits, frac_bits=frac_bits))
    val_input = (
        fxp.as_rational(
            torch.randint(
                low=fxp.config.minimum_as_integer,
                high=fxp.config.maximum_as_integer + 1,
                size=(features_in,),
            )
        )
        .int()
        .tolist()
    )

    output_dir = build_report_folder_and_testdata(
        dut_name=file_name,
        testdata={
            "data": val_input,
        },
    )

    # Build design and check if design is available
    output_dir_sub = Path(*output_dir.parts[len(find_project_root().parts) :])
    destination = OnDiskPath(str(output_dir_sub), parent=find_project_root())
    Rom(name=file_name, data_width=total_bits, values_as_integers=val_input).save_to(
        destination
    )
    assert exists(output_dir / f"{file_name}.vhd")
    assert exists(output_dir / "testdata.json")

    # Prepare cocotb runner and save waveforms
    run_cocotb_sim(
        src_files=[output_dir / f"{file_name}.vhd"],
        top_module_name=file_name,
        cocotb_test_module="tests.integration_tests.share_designs.rom.rom_tb",
        waveform_save_dst=str(output_dir),
    )



import pytest
import torch
from torch import nn as nn_torch

from elasticai.creator.arithmetic import FxpParams
from elasticai.creator.nn import fixed_point as nn_creator


@pytest.mark.parametrize(
    "total_bits, frac_bits, num_steps",
    [(4, 3, 4), (6, 4, 32), (8, 5, 64), (10, 6, 16), (12, 7, 128), (12, 9, 64)],
)
def test_silu_compared_torch(total_bits: int, frac_bits: int, num_steps: int) -> None:
    fxp = FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    vrange = (fxp.minimum_as_rational, fxp.maximum_as_rational)
    stimulus = torch.arange(
        start=vrange[0], end=vrange[1], step=fxp.minimum_step_as_rational
    )

    act0 = nn_torch.SiLU()
    out0 = act0(stimulus).detach().numpy()
    act1 = nn_creator.AdaptableSiLU(
        total_bits=total_bits,
        frac_bits=frac_bits,
        num_steps=num_steps,
        sampling_intervall=vrange,
    )
    out1 = act1(stimulus).detach().numpy()
    metric_mean_abs = float(sum(abs(out1 - out0))) / stimulus.shape[0]
    metric_mean = float(abs(sum(out1 - out0))) / stimulus.shape[0]

    assert metric_mean_abs < 2.2 * fxp.minimum_step_as_rational * (
        out0.max() - out0.min()
    )
    assert metric_mean < 0.6 * fxp.minimum_step_as_rational



from collections import OrderedDict
from typing import cast

import torch

from elasticai.creator.arithmetic import (
    FxpArithmetic,
    FxpParams,
)
from elasticai.creator.base_modules.lstm import LSTM
from elasticai.creator.base_modules.lstm_cell import LSTMCell
from elasticai.creator.nn.design_creator_module import DesignCreatorModule
from elasticai.creator.nn.fixed_point.hard_sigmoid import HardSigmoid
from elasticai.creator.nn.fixed_point.hard_tanh import HardTanh
from elasticai.creator.nn.fixed_point.lstm.design.fp_lstm_cell import FPLSTMCell
from elasticai.creator.nn.fixed_point.lstm.design.lstm import LSTMNetworkDesign
from elasticai.creator.vhdl.design.design import Design

from ..math_operations import MathOperations
from .design.testbench import LSTMTestBench


class LSTMNetwork(DesignCreatorModule):
    def __init__(self, layers: list[torch.nn.Module]):
        super().__init__()
        self.lstm = layers[0]
        self.layer_names = [f"fp_linear_{i}" for i in range(len(layers[1:]))]
        if len(self.layer_names) > 1:
            raise NotImplementedError
        self.layers = torch.nn.Sequential(
            OrderedDict(
                {name: layer for name, layer in zip(self.layer_names, layers[1:])}
            )
        )

    def create_design(self, name: str) -> LSTMNetworkDesign:
        first_lstm = cast(FixedPointLSTMWithHardActivations, self.lstm)
        total_bits = first_lstm.fixed_point_config.total_bits
        frac_bits = first_lstm.fixed_point_config.frac_bits
        hidden_size = first_lstm.hidden_size
        input_size = first_lstm.input_size

        return LSTMNetworkDesign(
            lstm=first_lstm.create_design(),
            linear_layer=self.layers[0].create_design(self.layer_names[0]),
            total_bits=total_bits,
            frac_bits=frac_bits,
            hidden_size=hidden_size,
            input_size=input_size,
        )

    def create_testbench(self, test_bench_name, uut: Design) -> LSTMTestBench:
        return LSTMTestBench(test_bench_name, uut)

    def forward(self, x):
        x, (_, _) = self.lstm(x)
        x = x[:, -1]
        return self.layers(x)


class FixedPointLSTMWithHardActivations(DesignCreatorModule, LSTM):
    """
    Use only together with the above `LSTMNetwork`.
    There is no single hw design corresponding to this sw layer.
    Instead, the design of the `LSTMNetwork` handles most of the tasks,
    that are performed by `FixedPointLSTMWithHardActivations`
    """

    def __init__(
        self,
        total_bits: int,
        frac_bits: int,
        input_size: int,
        hidden_size: int,
        bias: bool,
    ) -> None:
        params = FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
        config = FxpArithmetic(params)

        class LayerFactory:
            def lstm(self, input_size: int, hidden_size: int, bias: bool) -> LSTMCell:
                def activation(constructor):
                    def wrapped_constructor():
                        return constructor(total_bits=total_bits, frac_bits=frac_bits)

                    return wrapped_constructor

                return LSTMCell(
                    operations=MathOperations(config=config),
                    sigmoid_factory=activation(HardSigmoid),
                    tanh_factory=activation(HardTanh),
                    input_size=input_size,
                    hidden_size=hidden_size,
                    bias=bias,
                )

        super().__init__(
            input_size=input_size,
            hidden_size=hidden_size,
            bias=bias,
            batch_first=True,
            layers=LayerFactory(),
        )

        self._config = config

    @property
    def fixed_point_config(self) -> FxpArithmetic:
        return self._config

    def create_design(self, name: str = "lstm_cell") -> Design:
        def float_to_signed_int(value: float | list) -> int | list:
            if isinstance(value, list):
                return list(map(float_to_signed_int, value))
            return self._config.cut_as_integer(value)

        def cast_weights(x):
            return cast(list[list[list[int]]], x)

        def cast_bias(x):
            return cast(list[list[int]], x)

        return FPLSTMCell(
            name=name,
            hardtanh=self.cell.tanh.create_design(f"{name}_hardtanh"),
            hardsigmoid=self.cell.sigmoid.create_design(f"{name}_hardsigmoid"),
            total_bits=self._config.total_bits,
            frac_bits=self._config.frac_bits,
            w_ih=cast_weights(float_to_signed_int(self.cell.linear_ih.weight.tolist())),
            w_hh=cast_weights(float_to_signed_int(self.cell.linear_hh.weight.tolist())),
            b_ih=cast_bias(float_to_signed_int(self.cell.linear_ih.bias.tolist())),
            b_hh=cast_bias(float_to_signed_int(self.cell.linear_hh.bias.tolist())),
        )



from collections.abc import Callable, Iterable, Iterator
from typing import cast

from torch.fx import Node as FxNode
from torch.fx import Tracer
from torch.nn import Module

from elasticai.creator.function_utils import KeyedFunctionDispatcher
from elasticai.creator.graph import BaseGraph

from .core import Edge, Implementation, new_node
from .default_handlers import handlers as default_handlers


class _DefaultTracer(Tracer):
    def is_leaf_module(self, m, module_qualified_name):
        if type(m).__qualname__.startswith("elasticai"):
            return True
        return super().is_leaf_module(m, module_qualified_name)


class LoweringError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class Torch2Ir:
    def __init__(self, tracer: Tracer = _DefaultTracer()):
        super().__init__()
        self._tracer = tracer
        self._registry: dict[str, Implementation] = {}
        self._root = Implementation(graph=BaseGraph())
        self._root.type = "module"
        self._root.name = ""
        self._registry[""] = self._root
        self._extractors: KeyedFunctionDispatcher[Module, dict] = (
            KeyedFunctionDispatcher(self._get_module_key)
        )

    def register(
        self, module_type: str, handler: Callable[[Module], dict]
    ) -> Callable[[Module], dict]:
        """The handlers are used to extract the attributes of the module"""
        self._extractors.register(module_type, handler)
        return handler

    def register_handlers(
        self, handlers: Iterable[Callable[[Module], dict]]
    ) -> "Torch2Ir":
        for handler in handlers:
            self.register(handler.__name__, handler)
        return self

    @staticmethod
    def _get_module_key(module: Module) -> str:
        return module.__class__.__name__.lower()

    def convert(self, model: Module) -> Iterator[Implementation]:
        self.model = model
        torch_graph = self._tracer.trace(model)
        for node in torch_graph.nodes:
            self._handle_fx_node(node)
        registry = self._registry
        self._registry = {}
        yield from registry.values()

    def __call__(self, model: Module) -> Iterator[Implementation]:
        yield from self.convert(model)

    def _get_successors(self, node: FxNode) -> list[FxNode]:
        return list(node.users)

    def _get_type(self, node: FxNode) -> str:
        error = LoweringError("""
                              Cannot handle function calls or getting 
                              attributes during translation. Please use 
                              supported Modules to design your model
                              and change every function call to a
                            module call.""")
        match node.op:
            case "call_module":
                return type(
                    self.model.get_submodule(cast(str, node.target))
                ).__name__.lower()

            case "call_function":
                raise error
            case "placeholder":
                return "input"
            case "output":
                return "output"
            case "get_attr":
                raise error
            case _:
                raise Exception(f"Unknown node type: {node.op}")

    def _get_implementation(self, node: FxNode) -> str:
        match self._get_type(node):
            case "input":
                return "input"
            case "output":
                return "output"
            case _:
                return cast(str, node.target)

    def _handle_fx_node(self, node: FxNode) -> None:
        ir_node = new_node(
            name=node.name,
            type=self._get_type(node),
            implementation=self._get_implementation(node),
            attributes={},
        )
        self._root.add_node(ir_node)
        impl = ir_node.implementation
        if impl not in self._registry and impl not in ("input", "output"):
            self._registry[impl] = Implementation(
                name=impl,
                type=ir_node.type,
                graph=BaseGraph(),
                data=dict(**self._extract_attributes(node)),
            )

        for successor in self._get_successors(node):
            self._root.add_edge(Edge(src=node.name, dst=successor.name, data={}))

    def _extract_attributes(self, node: FxNode) -> dict:
        if self._get_type(node) in ("input", "output"):
            return {}
        module = self.model.get_submodule(cast(str, node.target))
        return self._extractors(module)


def get_default_converter() -> Torch2Ir:
    converter = Torch2Ir()
    converter.register_handlers(default_handlers)
    return converter



import pytest
import torch

import elasticai.creator.nn.fixed_point as nn_creator
from elasticai.creator.arithmetic import FxpArithmetic, FxpParams
from elasticai.creator.nn import Sequential
from elasticai.creator.nn.fixed_point.math_operations import (
    MathOperations,
)
from tests.integration_tests.nn.fixed_point.sequential_routine import (
    routine_testing_sequential_module,
)


@pytest.mark.simulation
@pytest.mark.parametrize(
    "total_bits, frac_bits, features_in, features_out",
    [
        (4, 2, 12, 6),
        (10, 8, 24, 20),
    ],
)
def test_build_test_linear_tanh(
    total_bits: int,
    frac_bits: int,
    features_in: int,
    features_out: int,
) -> None:
    file_name = f"TestLinearTanh_{total_bits}_{frac_bits}_{features_in}x{features_out}"
    fxp = FxpArithmetic(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    )
    math = MathOperations(fxp)

    dut = Sequential(
        nn_creator.Linear(
            in_features=features_in,
            out_features=features_out,
            total_bits=total_bits,
            frac_bits=frac_bits,
        ),
        nn_creator.Tanh(total_bits=total_bits, frac_bits=frac_bits, num_steps=16),
    )
    # --- Adapting values
    scale_amp = (fxp.maximum_as_rational - fxp.minimum_as_rational) / (2 * features_in)
    scale_min = -scale_amp / 2

    dut[0].weight.data = torch.nn.Parameter(
        math.quantize(
            scale_amp
            * torch.rand(
                size=(features_out, features_in),
            )
            + scale_min
            + torch.randint(low=-1, high=+1, size=(features_out, features_in))
            * fxp.config.minimum_step_as_rational
        )
    )
    dut[0].bias.data = torch.nn.Parameter(
        math.quantize(
            scale_amp
            * torch.rand(
                size=(features_out,),
            )
            + scale_min
            + torch.randint(low=-1, high=+1, size=(features_out,))
            * fxp.config.minimum_step_as_rational
        )
    )

    routine_testing_sequential_module(
        dut=dut,
        file_name=file_name,
        fxp=fxp,
        feat_in=features_in,
    )



from typing import Any

import pytest

from elasticai.creator.ir import (
    ReadOnlyField,
    RequiredField,
    read_only_field,
    static_required_field,
)
from elasticai.creator.ir.base.ir_data import IrData
from elasticai.creator.ir.base.ir_data_meta import IrDataMeta


def test_using_metaclass() -> None:
    class Node(metaclass=IrDataMeta, create_init=True):
        name: str

    n = Node(dict(name="x"))  # type: ignore
    assert "x" == n.name


def test_inheriting_with_metaclass() -> None:
    class Node(metaclass=IrDataMeta, create_init=True):
        name: str

    class MyNode(Node):
        type: str

    # __init__ not visible for type checker
    n = MyNode(dict(name="x", type="y"))  # type: ignore
    assert "y" == n.type


def test_inheriting_from_ir_base() -> None:
    class Node(IrData):
        name: str

    n = Node(dict(name="x"))
    assert "x" == n.name


def test_inherit_from_ir_base_child() -> None:
    class Node(IrData):
        name: str

    class Node2(Node):
        type: str

    n = Node2(data=dict(name="x", type="y"))
    assert "y" == n.type


def test_accidentally_writing_into_non_declared_field_raises_error() -> None:
    class Node(IrData):
        name: str

    n = Node(dict(name="x"))
    with pytest.raises(AttributeError):
        n.x = 5  # type: ignore


def test_inherited_class_contains_all_required_fields() -> None:
    class Node(IrData):
        name: str

    class Sub(Node):
        type: str

    assert dict(name=str, type=str).keys().mapping == Sub.required_fields


def test_get_missing_required_fields_returns_name_str() -> None:
    class Node(IrData):
        name: str

    class Sub(Node):
        type: str

    n = Sub(dict(type="a"))
    assert dict(name=str) == n.get_missing_required_fields()


def test_mandatory_fields_not_in_attributes() -> None:
    class Node(IrData):
        name: str
        type: str

    n = Node(dict(name="x", type="y", extra_field="f"))
    assert dict(extra_field="f") == n.attributes


def test_mandatory_fields_not_in_subclass_attributes() -> None:
    class Node(IrData):
        name: str

    class Sub(Node):
        type: str

    n = Sub(dict(name="a", type="b", extra="c"))
    assert dict(extra="c") == n.attributes


def test_defined_required_field_is_detected() -> None:
    class Node(IrData):
        width: RequiredField[str, int] = RequiredField(get_convert=int, set_convert=str)

    assert "width" in Node.required_fields


def test_defined_required_fields_type_is_detected() -> None:
    class Node(IrData):
        width: RequiredField[str, int] = RequiredField(get_convert=int, set_convert=str)

    assert issubclass(str, Node.required_fields["width"])


def test_can_increment_counter_saved_as_string() -> None:
    class Node(IrData):
        counter: RequiredField[str, int] = RequiredField(
            get_convert=int, set_convert=str
        )

    n = Node(dict(counter="1"))
    n.counter = n.counter + 2
    assert n.counter == 3
    assert n.data["counter"] == "3"


def test_read_only_field_is_detected() -> None:
    class Node(IrData):
        name: ReadOnlyField[str, str] = ReadOnlyField(lambda x: x)

    assert "name" in Node.required_fields


def test_attribute_is_read_only() -> None:
    class Node(IrData):
        name: ReadOnlyField[str, str] = ReadOnlyField(lambda x: x)

    n = Node(dict(name="x"))
    with pytest.raises(AttributeError):
        n.name = "y"


def test_right_oring_dict_with_attribute_does_not_add_hidden_fields() -> None:
    class Node(IrData):
        name: str
        type: str

    d: dict[str, Any] = dict()
    n = Node(dict(name="x", type="y", counter=1))
    d = d | n.attributes
    assert d == dict(counter=1)


def test_left_oring_dict_with_attribute_does_not_add_hidden_fields() -> None:
    class Node(IrData):
        name: str
        type: str

    d: dict[str, Any] = dict()
    n = Node(dict(name="x", type="y", counter=1))
    d = n.attributes | d
    assert d == dict(counter=1)


def test_inheriting_from_ir_data_without_annotated_fields_does_not_raise_exception():
    try:

        class Node(IrData):
            pass
    except KeyError:
        pytest.fail("did not expect an error")


class test_can_read_from_decorated_method_field:
    class Node(IrData):
        __slots__ = ("length_scaling",)

        def __init__(self, data: dict[str, Any]):
            super().__init__(data)
            self.length_scaling = 2

        @read_only_field
        def length(self, value: int) -> float:
            return value * self.length_scaling

    n = Node(dict(length=1))
    assert n.length == 2.0


class test_can_write_to_decorated_method_field:
    class Node(IrData):
        __slots__ = ("length_scaling",)

        def __init__(self, data: dict[str, Any]):
            super().__init__(data)

        @static_required_field
        @staticmethod
        def length(value: int) -> float:
            return value * 2

        @length.setter
        @staticmethod
        def _(value: float) -> int:
            return int(value / 2)

    n = Node(dict(length=1))
    n.length = 4.0
    assert n.data["length"] == 2



from unittest import TestCase

from torch import Tensor


def _unify_inputs(*inputs: float | list | Tensor) -> list[float | list]:
    def unify(obj: float | list | Tensor) -> float | list:
        if isinstance(obj, (float, list)):
            return obj
        return obj.detach().tolist()

    return list(map(unify, inputs))


class TensorTestCase(TestCase):
    def assertTensorEqual(
        self, expected: float | list | Tensor, actual: float | list | Tensor
    ) -> None:
        self.assertEqual(*_unify_inputs(expected, actual))


def assertTensorEqual(
    expected: float | list | Tensor, actual: float | list | Tensor
) -> None:
    expected, actual = _unify_inputs(expected, actual)
    assert expected == actual
